import numpy
import os

from AnyQt.QtWidgets import QLabel, QSizePolicy
from AnyQt.QtGui import QPixmap
from AnyQt.QtCore import Qt

from matplotlib import cm
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.figure import Figure

try:
    from mpl_toolkits.mplot3d import Axes3D  # necessario per caricare i plot 3D
except:
    pass

import orangecanvas.resources as resources
from orangewidget import gui
from orangewidget.settings import Setting

from oasys2.widget.gui import FigureCanvas3D, Styles
from oasys2.widget import gui as oasysgui
from oasys2.canvas.util.canvas_util import add_widget_parameters_to_module

from shadow4.beamline.optical_elements.mirrors.s4_ellipsoid_mirror import S4EllipsoidMirror
from shadow4_advanced.benders.s4_flexural_hinge_bender_ellipsoid_mirror import S4FlexuralHingeBenderEllipsoidMirror, S4FlexuralHingeBenderEllipsoidMirrorElement, \
    BenderType, MirrorShape, FlexuralHingeBenderFitParameters, BenderMovement, CalibrationParameters

from orangecontrib.shadow4.widgets.optics.ow_mirror import _OWMirror
from orangecontrib.shadow4.util.shadow4_objects import PreReflPreProcessorData
import copy

class OWFlexuralHingeBenderMirror(_OWMirror):
    name        = "Flexural Hinge Bender Mirror"
    description = "Flexural Hinge Bender Mirror"
    icon        = "icons/flexural_hinge_bender_mirror.png"

    priority = 1.1

    #########################################################
    # Bender
    #########################################################

    bender_bin_x     = Setting(100)
    bender_bin_y     = Setting(500)
    E                = Setting(131.0)
    h                = Setting(0.01)
    bender_type      = Setting(BenderType.DOUBLE_MOMENTUM)
    bender_shape     = Setting(MirrorShape.TRAPEZIUM)
    which_length     = Setting(0)
    optimized_length = Setting(0.0)
    n_fit_steps      = Setting(3)

    M1               = Setting(0.5)
    M2               = Setting(0.6)
    e                = Setting(0.3)

    M1_out           = 0.0
    M2_out        = 0.0
    e_out            = 0.0

    M1_fixed         = Setting(False)
    M2_fixed         = Setting(False)
    e_fixed          = Setting(False)

    M1_min           = Setting(0.0)
    M2_min           = Setting(0.0)
    e_min            = Setting(0.0)

    M1_max           = Setting(1000.0)
    M2_max           = Setting(10.0)
    e_max            = Setting(1.0)

    use_fitted_result_to_move_bender = Setting(0)
    use_calibration                  = Setting(0)

    q_downstream  = Setting(0.0)
    q_upstream    = Setting(0.0)

    p0_downstream = Setting(0.0)
    p1_downstream = Setting(0.0)
    p0_upstream   = Setting(0.0)
    p1_upstream   = Setting(0.0)

    pos_downstream = Setting(0.0)
    pos_upstream   = Setting(0.0)

    q_downstream_movement = 0.0
    q_upstream_movement   = 0.0

    M1_movement = 0.0
    M2_movement = 0.0

    show_bender_plots = Setting(0)

    help_path        = os.path.join(resources.package_dirname("orangecontrib.shadow4_advanced.widgets.tools"), "icons", "flexural_hinge_bender_scheme.png")
    calibration_path = os.path.join(resources.package_dirname("orangecontrib.shadow4_advanced.widgets.tools"), "icons", "calibration.png")


    def __init__(self):
        super(OWFlexuralHingeBenderMirror, self).__init__(switch_icons=False)

        # FIXED SHAPE
        self.surface_shape_type       = 2 # Ellipsoid
        self.surface_shape_parameters = 0 # Internal
        self.surface_shape_tab_visibility(is_init=True)
        self.surface_shape_type_combo.setEnabled(False)
        self.surface_shape_parameters_combo.setEnabled(False)

        # FINITE DIMENSIONS
        self.is_infinite = 0
        self.dimensions_tab_visibility()
        self.is_infinite_combo.setEnabled(False)

        plot_tab = oasysgui.createTabPage(self.main_tabs, "Bender Plots")

        view_box = oasysgui.widgetBox(plot_tab, "Plotting Style", addSpace=False, orientation="vertical", width=350)

        self.view_type_combo = gui.comboBox(view_box, self, "show_bender_plots", label="Show Plots", labelWidth=220,
                                            items=["No", "Yes"], sendSelectedValue=False, orientation="horizontal")

        bender_tabs = oasysgui.tabWidget(plot_tab)

        tabs = [oasysgui.createTabPage(bender_tabs, "Bender vs. Ideal (1D)"),
                oasysgui.createTabPage(bender_tabs, "Ideal - Bender (1D)"),
                oasysgui.createTabPage(bender_tabs, "Ideal - Bender (3D)"),
                oasysgui.createTabPage(bender_tabs, "Figure Error (3D)"),
                oasysgui.createTabPage(bender_tabs, "Ideal - Bender + Figure Error (3D)")]

        def create_figure_canvas(mode="3D"):
            figure = Figure(figsize=(100, 100))
            figure.patch.set_facecolor('white')
            if mode == "3D":
                ax = figure.add_subplot(111, projection='3d')
                figure_canvas = FigureCanvas3D(ax=ax, fig=figure)
            else:
                figure.add_subplot(111)
                figure_canvas = FigureCanvasQTAgg(figure)
            figure_canvas.setFixedWidth(self.IMAGE_WIDTH)
            figure_canvas.setFixedHeight(self.IMAGE_HEIGHT-10)

            return figure_canvas

        self.figure_canvas = [create_figure_canvas("1D"), create_figure_canvas("1D"),
                              create_figure_canvas("3D"), create_figure_canvas("3D"), create_figure_canvas("3D")]

        for tab, figure_canvas in zip(tabs, self.figure_canvas): tab.layout().addWidget(figure_canvas)

        gui.rubber(self.controlArea)
        gui.rubber(self.mainArea)

    def create_basic_settings_specific_subtabs(self, tabs_basic_setting): 
        specific_subtabs = [super(OWFlexuralHingeBenderMirror, self).create_basic_settings_specific_subtabs(tabs_basic_setting),
                            oasysgui.createTabPage(tabs_basic_setting, "Bender")]
        
        return specific_subtabs

    def populate_basic_settings_specific_subtabs(self, specific_subtabs):
        super(OWFlexuralHingeBenderMirror, self).populate_basic_settings_specific_subtabs(specific_subtabs[0])
        
        subtab_bender = specific_subtabs[1]

        #########################################################
        # Bender
        #########################################################
        self.populate_tab_bender(subtab_bender)
    
    def populate_tab_bender(self, subtab_bender):
        tabs = gui.tabWidget(subtab_bender)

        tab_bender = oasysgui.createTabPage(tabs, "Bender")

        surface_box = oasysgui.widgetBox(tab_bender, "Surface", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(surface_box, self, "bender_bin_x", "bins Sagittal", labelWidth=260, valueType=int, orientation="horizontal")
        oasysgui.lineEdit(surface_box, self, "bender_bin_y", "bins Transversal", labelWidth=260, valueType=int, orientation="horizontal")

        material_box = oasysgui.widgetBox(tab_bender, "Bender", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(material_box, self, "E", "Young's Modulus [GPa]", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(material_box, self, "h", "Thickness [m]", labelWidth=260, valueType=float, orientation="horizontal")

        gui.comboBox(material_box, self, "bender_type", label="Kind Of Bender ", items=["Single Momentum", "Double Momentum"],
                     labelWidth=150, orientation="horizontal", callback=self.set_bender_type)

        gui.comboBox(material_box, self, "bender_shape", label="Mirror Shape ", items=["Trapezium", "Rectangle"],
                     labelWidth=150, orientation="horizontal", callback=self.set_bender_shape)

        help_box = oasysgui.widgetBox(tab_bender, "", addSpace=False, orientation="horizontal")

        label = QLabel("")
        label.setAlignment(Qt.AlignCenter | Qt.AlignTop)
        label.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        label.setPixmap(QPixmap(self.help_path).scaledToWidth(340))

        help_box.layout().addWidget(label)


        tab_fit = oasysgui.createTabPage(tabs, "Fit")

        self.fit_box = oasysgui.widgetBox(tab_fit, "", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(self.fit_box, self, "n_fit_steps", "Nr. fit steps", labelWidth=250, valueType=int, orientation="horizontal")

        length_box = oasysgui.widgetBox(self.fit_box, "", addSpace=False, orientation="horizontal")

        self.cb_optimized_length = gui.comboBox(length_box, self, "which_length", label="Optimized Length ", items=["Total", "Partial"],
                                                labelWidth=150, orientation="horizontal", callback=self.set_which_length)
        self.le_optimized_length = oasysgui.lineEdit(length_box, self, "optimized_length", " ", labelWidth=10, valueType=float, orientation="horizontal")
        self.set_which_length()

        gui.separator(self.fit_box)

        def add_parameter_box(container_box, variable, label):
            box = oasysgui.widgetBox(container_box, "", addSpace=False, orientation="horizontal")
            oasysgui.lineEdit(box, self, variable, label, labelWidth=60, valueType=float, orientation="horizontal")
            gui.label(box, self, " ", labelWidth=58)

            box = oasysgui.widgetBox(container_box, "", addSpace=False, orientation="horizontal")

            setattr(self, "le_" + variable + "_min", oasysgui.lineEdit(box, self, variable + "_min", "Min", labelWidth=60, valueType=float, orientation="horizontal"))
            setattr(self, "le_" + variable + "_max", oasysgui.lineEdit(box, self, variable + "_max", "Max", labelWidth=35, valueType=float, orientation="horizontal"))

            gui.checkBox(box, self, variable + "_fixed", "Fixed", callback=getattr(self, "set_" + variable))

            box = oasysgui.widgetBox(container_box, "", addSpace=False, orientation="horizontal")

            le = oasysgui.lineEdit(box, self, variable + "_out", "Fitted", labelWidth=60, valueType=float, orientation="horizontal")
            le.setEnabled(False)
            le.setStyleSheet(Styles.line_edit_read_only)

            def set_variable_fit(): setattr(self, variable, getattr(self, variable + "_out"))
            gui.button(box, self, "<- Use", width=58, callback=set_variable_fit)

            getattr(self, "set_" + variable)()

        self.M2_box = oasysgui.widgetBox(self.fit_box, "", addSpace=False, orientation="vertical")
        gui.separator(self.fit_box, 10)
        m1_box = oasysgui.widgetBox(self.fit_box, "", addSpace=False, orientation="vertical")
        gui.separator(self.fit_box, 10)
        self.e_box = oasysgui.widgetBox(self.fit_box, "", addSpace=False, orientation="vertical")
        gui.separator(self.fit_box, 10)

        add_parameter_box(m1_box, "M1", "Md [Nm]")
        add_parameter_box(self.M2_box, "M2", "Mu [Nm]")
        add_parameter_box(self.e_box, "e", "e")

        box = oasysgui.widgetBox(tab_fit, "", addSpace=False, orientation="vertical")

        gui.comboBox(box, self, "use_fitted_result_to_move_bender", label="Use Fit Result to move the Bender ", items=["No", "Yes"],
                     labelWidth=250, orientation="horizontal", callback=self.set_use_fitted_result_to_move_bender)

        self.tab_mov = oasysgui.createTabPage(tabs, "Movements")

        gui.comboBox(self.tab_mov, self, "use_calibration", label="Use calibration to move the Bender ", items=["No", "Yes"],
                     labelWidth=250, orientation="horizontal", callback=self.set_use_calibration)

        self.mov_box_nocal = oasysgui.widgetBox(self.tab_mov, "", addSpace=False, orientation="vertical")
        self.mov_box_cal   = oasysgui.widgetBox(self.tab_mov, "", addSpace=False, orientation="vertical")

        # ---- NO CALIBRATION

        self.mov_box_nocal_1 = oasysgui.widgetBox(self.mov_box_nocal, "Double Momentum", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(self.mov_box_nocal_1, self, "q_upstream", "Upstream Focus Position [m]", labelWidth=250, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.mov_box_nocal_1, self, "q_downstream", "Downstream Focus Position [m]", labelWidth=250, valueType=float, orientation="horizontal")

        le = oasysgui.lineEdit(self.mov_box_nocal_1, self, "M2_movement", "Mu [Nm]", labelWidth=250, valueType=float, orientation="horizontal")
        le.setEnabled(False)
        le.setStyleSheet(Styles.line_edit_read_only)
        le = oasysgui.lineEdit(self.mov_box_nocal_1, self, "M1_movement", "Md [Nm]", labelWidth=250, valueType=float, orientation="horizontal")
        le.setEnabled(False)
        le.setStyleSheet(Styles.line_edit_read_only)

        self.mov_box_nocal_2 = oasysgui.widgetBox(self.mov_box_nocal, "Single Momentum", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(self.mov_box_nocal_2, self, "q_downstream", "Focus Position [m]", labelWidth=250, valueType=float, orientation="horizontal")

        le = oasysgui.lineEdit(self.mov_box_nocal_2, self, "M1_movement", "Md [Nm]", labelWidth=250, valueType=float, orientation="horizontal")
        le.setEnabled(False)
        le.setStyleSheet(Styles.line_edit_read_only)

        # ---- CALIBRATION

        self.mov_box_cal_1 = oasysgui.widgetBox(self.mov_box_cal, "Double Momentum", addSpace=False, orientation="vertical")

        box = oasysgui.widgetBox(self.mov_box_cal_1, "", addSpace=False, orientation="horizontal")
        oasysgui.lineEdit(box, self, "p0_upstream",  "Upstream 1/q = P0", labelWidth=130, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box, self, "p1_upstream", "\u00b7pos + P1", labelWidth=60, valueType=float, orientation="horizontal")

        box = oasysgui.widgetBox(self.mov_box_cal_1, "", addSpace=False, orientation="horizontal")
        oasysgui.lineEdit(box, self, "p0_downstream",  "Downstream 1/q = P0", labelWidth=130, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box, self, "p1_downstream", "\u00b7pos + P1", labelWidth=60, valueType=float, orientation="horizontal")

        oasysgui.lineEdit(self.mov_box_cal_1, self, "pos_upstream", "Upstream Motor", labelWidth=250, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.mov_box_cal_1, self, "pos_downstream", "Downstream Motor", labelWidth=250, valueType=float, orientation="horizontal")

        le = oasysgui.lineEdit(self.mov_box_cal_1, self, "M2_movement", "Mu [Nm]", labelWidth=250, valueType=float, orientation="horizontal")
        le.setEnabled(False)
        le.setStyleSheet(Styles.line_edit_read_only)
        le = oasysgui.lineEdit(self.mov_box_cal_1, self, "M1_movement", "Md [Nm]", labelWidth=250, valueType=float, orientation="horizontal")
        le.setEnabled(False)
        le.setStyleSheet(Styles.line_edit_read_only)

        le = oasysgui.lineEdit(self.mov_box_cal_1, self, "q_upstream_movement", "Upstream Focus Position [m]", labelWidth=250, valueType=float, orientation="horizontal")
        le.setEnabled(False)
        le.setStyleSheet(Styles.line_edit_read_only)
        le = oasysgui.lineEdit(self.mov_box_cal_1, self, "q_downstream_movement", "Downstream Focus Position [m]", labelWidth=250, valueType=float, orientation="horizontal")
        le.setEnabled(False)
        le.setStyleSheet(Styles.line_edit_read_only)

        self.mov_box_cal_2 = oasysgui.widgetBox(self.mov_box_cal, "Single Momentum", addSpace=False, orientation="vertical")

        box = oasysgui.widgetBox(self.mov_box_cal_2, "", addSpace=False, orientation="horizontal")
        oasysgui.lineEdit(box, self, "p0_downstream",  "1/q = P0", labelWidth=130, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box, self, "p1_downstream", "\u00b7pos + P1", labelWidth=60, valueType=float, orientation="horizontal")

        oasysgui.lineEdit(self.mov_box_cal_2, self, "pos_downstream", "Bender Motor", labelWidth=250, valueType=float, orientation="horizontal")

        le = oasysgui.lineEdit(self.mov_box_cal_2, self, "M1_movement", "Md/u [Nm]", labelWidth=250, valueType=float, orientation="horizontal")
        le.setEnabled(False)
        le.setStyleSheet(Styles.line_edit_read_only)

        le = oasysgui.lineEdit(self.mov_box_cal_2, self, "q_downstream_movement", "Focus Position [m]", labelWidth=250, valueType=float, orientation="horizontal")
        le.setEnabled(False)
        le.setStyleSheet(Styles.line_edit_read_only)

        self.set_bender_shape()
        self.set_bender_type()
        self.set_use_fitted_result_to_move_bender()

    def set_use_calibration(self):
        self.mov_box_nocal.setVisible(self.use_calibration==0)
        self.mov_box_cal.setVisible(self.use_calibration==1)

    def set_use_fitted_result_to_move_bender(self):
        self.fit_box.setEnabled(self.use_fitted_result_to_move_bender==0)
        self.tab_mov.setEnabled(self.use_fitted_result_to_move_bender==1)

    def set_bender_type(self):
        self.M2_box.setVisible(self.bender_type == BenderType.DOUBLE_MOMENTUM)
        self.mov_box_nocal_1.setVisible(self.bender_type == BenderType.DOUBLE_MOMENTUM)
        self.mov_box_nocal_2.setVisible(self.bender_type == BenderType.SINGLE_MOMENTUM)
        self.mov_box_cal_1.setVisible(self.bender_type == BenderType.DOUBLE_MOMENTUM)
        self.mov_box_cal_2.setVisible(self.bender_type == BenderType.SINGLE_MOMENTUM)
        self.set_use_calibration()

    def set_bender_shape(self):
        self.e_box.setVisible(self.bender_shape == MirrorShape.TRAPEZIUM)

    def set_which_length(self):
        self.le_optimized_length.setEnabled(self.which_length == 1)

    def set_M1(self):
        self.le_M1_min.setEnabled(self.M1_fixed == False)
        self.le_M1_max.setEnabled(self.M1_fixed == False)

    def set_M2(self):
        self.le_M2_min.setEnabled(self.M2_fixed == False)
        self.le_M2_max.setEnabled(self.M2_fixed == False)

    def set_e(self):
        self.le_e_min.setEnabled(self.e_fixed == False)
        self.le_e_max.setEnabled(self.e_fixed == False)

    #########################################################
    # S4 objects
    #########################################################

    def _post_trace_operations(self, output_beam, footprint, element, beamline):
        bender_data = element.get_optical_element().get_bender_data()

        if self.use_fitted_result_to_move_bender == 0:
            self.M1_out = round(bender_data.M1_out, 4)
            if self.bender_shape==MirrorShape.TRAPEZIUM:     self.e_out     = round(bender_data.e_out, 5)
            if self.bender_type==BenderType.DOUBLE_MOMENTUM: self.M2_out    = round(bender_data.M1_out * bender_data.ratio_out, 4)
        else:
            self.M1_movement = round(bender_data.M1_out, 4)
            if self.bender_type==BenderType.DOUBLE_MOMENTUM: self.M2_movement = round(bender_data.M1_out * bender_data.ratio_out, 4)

            if self.use_calibration==1:
                def get_q(pos, p0, p1): return round(1/(p0*pos + p1), 6)

                self.q_downstream_movement = get_q(self.pos_downstream, self.p0_downstream, self.p1_downstream)
                if self.bender_type == BenderType.DOUBLE_MOMENTUM: self.q_upstream_movement = get_q(self.pos_upstream, self.p0_upstream, self.p1_upstream)

    def _plot_additional_results(self, output_beam, footprint, element, beamline):
        if self.show_bender_plots == 1:
            bender_data = element.get_optical_element().get_bender_data()

            self.plot1D(bender_data.y, bender_data.bender_profile, y_values_2=bender_data.ideal_profile,
                        index=0, title=bender_data.titles[0], um=1)
            self.plot1D(bender_data.y, bender_data.correction_profile,
                        index=1, title=bender_data.titles[1])

            self.plot3D(bender_data.x,
                        bender_data.y,
                        bender_data.z_bender_correction_no_figure_error,
                        index=2, title="Ideal - Bender Surfaces")

            if self.modified_surface > 0:
                self.plot3D(bender_data.x,
                            bender_data.y,
                            bender_data.z_figure_error,  index=3, title="Figure Error Surface")
                self.plot3D(bender_data.x,
                            bender_data.y,
                            bender_data.z_bender_correction, index=4, title="Ideal - Bender + Figure Error Surfaces")


    def get_optical_element_instance(self):
        if self.modified_surface: self.congruence_surface_data_file()

        reflectivity_source = self.reflectivity_source if self.reflectivity_source < 5 else 6 # no more xraylib

        ellipsoid_mirror = S4EllipsoidMirror(
                name="Flexural Hinge Bender Mirror",
                boundary_shape=self.get_boundary_shape(),
                surface_calculation=self.surface_shape_parameters, # INTERNAL = 0  EXTERNAL = 1
                is_cylinder=self.is_cylinder,
                cylinder_direction=self.cylinder_orientation, #  Direction:  TANGENTIAL = 0  SAGITTAL = 1
                convexity=numpy.logical_not(self.surface_curvature).astype(int), #  Convexity: NONE = -1  UPWARD = 0  DOWNWARD = 1
                min_axis=self.ellipse_hyperbola_semi_minor_axis * 2, # todo: check factor 2
                maj_axis=self.ellipse_hyperbola_semi_major_axis * 2, # todo: check factor 2
                pole_to_focus=self.angle_of_majax_and_pole, # todo: change variable name
                p_focus=self.get_focusing_p(),
                q_focus=self.get_focusing_q(),
                grazing_angle=self.get_focusing_grazing_angle(),
                # inputs related to mirror reflectivity
                f_reflec=self.reflectivity_flag,  # reflectivity of surface: 0=no reflectivity, 1=full polarization
                f_refl=reflectivity_source,  # 0=prerefl file
                                             # 1=electric susceptibility
                                             # 2=user defined file (1D reflectivity vs angle)
                                             # 3=user defined file (1D reflectivity vs energy)
                                             # 4=user defined file (2D reflectivity vs energy and angle)
                                             # 5=direct calculation using xraylib
                                             # 6=direct calculation using dabax
                file_refl=self.file_refl,  # preprocessor file fir f_refl=0,2,3,4
                refraction_index=1 - self.refraction_index_delta + 1j * self.refraction_index_beta,
                # refraction index (complex) for f_refl=1
                coating_material=self.coating_material,    # string with coating material formula for f_refl=5,6
                coating_density=self.coating_density,      # coating material density for f_refl=5,6
                coating_roughness=self.coating_roughness,  # coating material roughness in A for f_refl=5,6
                )

        if self.use_fitted_result_to_move_bender == 0:
            fit_to_focus_parameters = FlexuralHingeBenderFitParameters(optimized_length = self.optimized_length if self.which_length==1 else None,
                                                                       n_fit_steps      = self.n_fit_steps,
                                                                       M1               = self.M1,
                                                                       M1_min           = self.M1_min,
                                                                       M1_max           = self.M1_max,
                                                                       M1_fixed         = self.M1_fixed,
                                                                       e                = self.e,
                                                                       e_min            = self.e_min,
                                                                       e_max            = self.e_max,
                                                                       e_fixed          = self.e_fixed,
                                                                       ratio            = (self.M2 / self.M1),
                                                                       ratio_min        = (self.M2_min / self.M1),
                                                                       ratio_max        = (self.M2_max / self.M1),
                                                                       ratio_fixed      = self.M2_fixed)

            return S4FlexuralHingeBenderEllipsoidMirror(ellipsoid_mirror=ellipsoid_mirror,
                                                        figure_error_data_file=self.ms_defect_file_name if self.modified_surface else None,
                                                        bender_bin_x=self.bender_bin_x,
                                                        bender_bin_y=self.bender_bin_y,
                                                        E=self.E*1e9,
                                                        h=self.h,
                                                        bender_shape=self.bender_shape,
                                                        bender_type=self.bender_type,
                                                        fit_to_focus_parameters=fit_to_focus_parameters)
        else:
            if self.use_calibration == 0:
                bender_movement = BenderMovement(position_upstream=self.q_upstream, position_downstream=self.q_downstream)

                return S4FlexuralHingeBenderEllipsoidMirror(ellipsoid_mirror=ellipsoid_mirror,
                                                            figure_error_data_file=self.ms_defect_file_name if self.modified_surface else None,
                                                            bender_bin_x=self.bender_bin_x,
                                                            bender_bin_y=self.bender_bin_y,
                                                            E=self.E*1e9,
                                                            h=self.h,
                                                            M1=self.M1,
                                                            e=self.e,
                                                            ratio=self.M2 / self.M1,
                                                            bender_shape=self.bender_shape,
                                                            bender_type=self.bender_type,
                                                            bender_movement=bender_movement)
            else:
                calibration_parameters = CalibrationParameters(parameters_upstream=[self.p0_upstream, self.p1_upstream],
                                                               parameters_downstream=[self.p0_downstream, self.p1_downstream])

                bender_movement = BenderMovement(position_upstream=self.pos_upstream, position_downstream=self.pos_downstream)

                return S4FlexuralHingeBenderEllipsoidMirror(ellipsoid_mirror=ellipsoid_mirror,
                                                            figure_error_data_file=self.ms_defect_file_name if self.modified_surface else None,
                                                            bender_bin_x=self.bender_bin_x,
                                                            bender_bin_y=self.bender_bin_y,
                                                            E=self.E*1e9,
                                                            h=self.h,
                                                            M1=self.M1,
                                                            e=self.e,
                                                            ratio=self.M2 / self.M1,
                                                            bender_shape=self.bender_shape,
                                                            bender_type=self.bender_type,
                                                            calibration_parameters=calibration_parameters,
                                                            bender_movement=bender_movement)

    def get_beamline_element_instance(self): return S4FlexuralHingeBenderEllipsoidMirrorElement()

    def plot1D(self, x_coords, y_values, y_values_2=None, index=0, title="", um=0):
        if self.show_bender_plots == 1:
            figure = self.figure_canvas[index].figure

            axis = figure.gca()
            axis.clear()

            axis.set_xlabel("Y [mm]")
            axis.set_ylabel("Z [" + ("nm" if um == 0 else "\u03bcm") + "]")
            axis.set_title(title)

            axis.plot(x_coords * 1e3, y_values * (1e9 if um == 0 else 1e6), color="blue", label="bender", linewidth=2)
            if not y_values_2 is None: axis.plot(x_coords * 1e3, y_values_2 * (1e9 if um == 0 else 1e6), "-.r", label="ideal")

            axis.legend(loc=0, fontsize='small')

            figure.canvas.draw()

    def plot3D(self, x_coords, y_coords, z_values, index, title=""):
        if self.show_bender_plots == 1:
            figure = self.figure_canvas[index].figure
            x_to_plot, y_to_plot = numpy.meshgrid(x_coords, y_coords)
            z_to_plot = z_values.T

            axis = figure.gca()
            axis.clear()

            axis.set_xlabel("X [mm]")
            axis.set_ylabel("Y [mm]")
            axis.set_zlabel("Z [nm]")
            axis.set_title(title)

            axis.plot_surface(x_to_plot * 1e3,
                              y_to_plot * 1e3 ,
                              z_to_plot * 1e9,
                              rstride=1, cstride=1, cmap=cm.autumn, linewidth=0.5, antialiased=True)

            figure.canvas.draw()

            axis.mouse_init()

add_widget_parameters_to_module(__name__)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# #########################################################################
# Copyright (c) 2020, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2020. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# #########################################################################
import os
import sys
import numpy

from silx.gui.plot import Plot2D

from AnyQt.QtWidgets import QMessageBox, QDialog, QVBoxLayout, QLabel, QDialogButtonBox
from AnyQt.QtGui import QPixmap, QPalette, QColor, QFont

import orangecanvas.resources as resources

from orangewidget import gui
from orangewidget.settings import Setting

from oasys2.widget import gui as oasysgui
from oasys2.widget.util import congruence
from oasys2.widget.util.widget_util import EmittingStream
from oasys2.widget.util.widget_objects import TriggerIn
from oasys2.canvas.util.canvas_util import add_widget_parameters_to_module

from syned.beamline.optical_elements.absorbers.slit import Slit
from syned.beamline.shape import Rectangle

try:
    from orangecontrib.shadow4.widgets.gui.ow_synchrotron_source import OWSynchrotronSource
    from orangecontrib.shadow4.util.shadow4_objects import ShadowData
    from orangecontrib.shadow4.util.shadow4_util import ShadowCongruence
except ImportError:
    pass

from shadow4.beamline.s4_beamline import S4Beamline

from shadow4_advanced.hybrid.s4_hybrid_undulator_light_source import (
    S4HybridUndulatorLightSource, HybridUndulatorInputParameters, HybridUndulatorOutputParameters, HybridUndulatorListener,
    gamma, get_default_initial_z, resonance_energy, is_canted_undulator, get_source_slit_data, set_which_waist, K_from_magnetic_field, S4HybridUndulator
)

import scipy.constants as codata

m2ev = codata.c * codata.h / codata.e

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure

class AutoUndulator:
    VERTICAL   = 1
    HORIZONTAL = 2
    BOTH       = 3

class HybridUndulator(OWSynchrotronSource, HybridUndulatorListener):
    name = "Shadow4/SRW Undulator"
    description = "Advanced Tools: Hybrid Shadow/SRW Undulator"
    icon = "icons/hybrid_undulator.png"
    priority = 5
    maintainer = "Luca Rebuffi"
    maintainer_email = "lrebuffi(@at@)anl.gov"
    category = "Sources"
    keywords = ["data", "file", "load", "read"]

    distribution_source = Setting(0)
    cumulated_view_type = Setting(0)

    # Shadow
    number_of_rays  = Setting(5000)
    seed            = Setting(6775431)
    use_harmonic    = Setting(0)
    harmonic_number = Setting(1)
    harmonic_energy = 0.0
    energy          = Setting(10000.0)
    energy_to       = Setting(10100.0)
    energy_points   = Setting(10)

    #SRW -> Undulator
    number_of_periods                            = Setting(184)
    undulator_period                             = Setting(0.025)
    Kv                                           = Setting(0.857)
    Kh                                           = Setting(0)
    Bh                                           = Setting(0.0)
    Bv                                           = Setting(1.5)
    magnetic_field_from                          = Setting(0)
    initial_phase_vertical                       = Setting(0.0)
    initial_phase_horizontal                     = Setting(0.0)
    symmetry_vs_longitudinal_position_vertical   = Setting(1)
    symmetry_vs_longitudinal_position_horizontal = Setting(0)
    horizontal_central_position                  = Setting(0.0)
    vertical_central_position                    = Setting(0.0)
    longitudinal_central_position                = Setting(0.0)

    auto_expand = Setting(0)
    auto_expand_rays = Setting(0)

    type_of_initialization = Setting(0)

    moment_x = Setting(0.0)
    moment_y = Setting(0.0)
    moment_z = Setting(0.0)
    moment_xp = Setting(0.0)
    moment_yp = Setting(0.0)

    source_dimension_wf_h_slit_gap = Setting(0.0015)
    source_dimension_wf_v_slit_gap = Setting(0.0015)
    source_dimension_wf_h_slit_c = Setting(0.0)
    source_dimension_wf_v_slit_c =Setting( 0.0)
    source_dimension_wf_h_slit_points = Setting(301)
    source_dimension_wf_v_slit_points = Setting(301)
    source_dimension_wf_distance = Setting(28.0)

    horizontal_range_modification_factor_at_resizing = Setting(0.5)
    horizontal_resolution_modification_factor_at_resizing = Setting(5.0)
    vertical_range_modification_factor_at_resizing = Setting(0.5)
    vertical_resolution_modification_factor_at_resizing = Setting(5.0)

    waist_position_calculation = Setting(0)
    waist_position = Setting(0.0)

    waist_position_auto = Setting(0)
    waist_position_auto_h = Setting(0.0)
    waist_position_auto_v = Setting(0.0)
    waist_back_propagation_parameters = Setting(1)
    waist_horizontal_range_modification_factor_at_resizing = Setting(0.5)
    waist_horizontal_resolution_modification_factor_at_resizing = Setting(5.0)
    waist_vertical_range_modification_factor_at_resizing = Setting(0.5)
    waist_vertical_resolution_modification_factor_at_resizing = Setting(5.0)
    which_waist = Setting(2)
    number_of_waist_fit_points = Setting(10)
    degree_of_waist_fit = Setting(3)
    use_sigma_or_fwhm = Setting(0)

    waist_position_user_defined = Setting(0.0)

    kind_of_sampler = Setting(1)
    save_srw_result = Setting(0)

    # SRW FILE INPUT

    source_dimension_srw_file = Setting("intensity_source_dimension.dat")
    angular_distribution_srw_file = Setting("intensity_angular_distribution.dat")

    # ASCII FILE INPUT

    x_positions_file = Setting("x_positions.txt")
    z_positions_file = Setting("z_positions.txt")
    x_positions_factor = Setting(0.01)
    z_positions_factor = Setting(0.01)
    x_divergences_file = Setting("x_divergences.txt")
    z_divergences_file = Setting("z_divergences.txt")
    x_divergences_factor = Setting(1.0)
    z_divergences_factor = Setting(1.0)

    combine_strategy = Setting(0)

    # Utility

    auto_energy          = Setting(0.0)
    auto_harmonic_number = Setting(1)

    # Advanced
    use_stokes = Setting(1)

    energy_step     = None
    power_step      = None
    current_step    = None
    total_steps     = None
    start_event     = True
    compute_power   = False
    integrated_flux = None
    power_density   = None

    cumulated_energies = None
    cumulated_integrated_flux = None
    cumulated_power_density = None
    cumulated_power = None

    do_cumulated_calculations = False

    def __init__(self):
        super().__init__(show_energy_spread=True)

        tab_shadow = oasysgui.createTabPage(self.tabs_control_area, "Shadow4")
        tab_spdiv  = oasysgui.createTabPage(self.tabs_control_area, "Position/Divergence")
        tab_util   = oasysgui.createTabPage(self.tabs_control_area, "Utility")

        ####################################################################################
        # SHADOW

        left_box_1 = oasysgui.widgetBox(tab_shadow, "Monte Carlo and Energy Spectrum", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(left_box_1, self, "number_of_rays", "Number of Rays", tooltip="Number of Rays", labelWidth=250, valueType=int, orientation="horizontal")
        oasysgui.lineEdit(left_box_1, self, "seed", "Seed", tooltip="Seed (0=clock)", labelWidth=250, valueType=int, orientation="horizontal")

        gui.comboBox(left_box_1, self, "use_harmonic", label="Photon Energy",
                     items=["Harmonic", "Other", "Range"], labelWidth=260,
                     callback=self.set_wf_use_harmonic, sendSelectedValue=False, orientation="horizontal")

        self.use_harmonic_box_1 = oasysgui.widgetBox(left_box_1, "", addSpace=False, orientation="vertical", height=80)
        oasysgui.lineEdit(self.use_harmonic_box_1, self, "harmonic_number", "Harmonic #", labelWidth=260, valueType=int, orientation="horizontal", callback=self.set_harmonic_energy)
        le_he = oasysgui.lineEdit(self.use_harmonic_box_1, self, "harmonic_energy", "Harmonic Energy", labelWidth=260, valueType=float, orientation="horizontal")
        le_he.setReadOnly(True)
        font = QFont(le_he.font())
        font.setBold(True)
        le_he.setFont(font)
        palette = QPalette(le_he.palette())
        palette.setColor(QPalette.Text, QColor('dark blue'))
        palette.setColor(QPalette.Base, QColor(243, 240, 160))
        le_he.setPalette(palette)

        self.use_harmonic_box_2 = oasysgui.widgetBox(left_box_1, "", addSpace=False, orientation="vertical", height=80)
        oasysgui.lineEdit(self.use_harmonic_box_2, self, "energy", "Photon Energy [eV]", labelWidth=260, valueType=float, orientation="horizontal")

        self.use_harmonic_box_3 = oasysgui.widgetBox(left_box_1, "", addSpace=False, orientation="vertical", height=80)
        oasysgui.lineEdit(self.use_harmonic_box_3, self, "energy", "Photon Energy from [eV]", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.use_harmonic_box_3, self, "energy_to", "Photon Energy to [eV]", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.use_harmonic_box_3, self, "energy_points", "Nr. of Energy values", labelWidth=260, valueType=int, orientation="horizontal")

        self.set_wf_use_harmonic()

        gui.comboBox(tab_spdiv, self, "distribution_source", label="Distribution Source", labelWidth=310,
                     items=["SRW Calculation", "SRW Files", "ASCII Files"], orientation="horizontal", callback=self.set_distribution_source)

        self.srw_box       = oasysgui.widgetBox(tab_spdiv, "", addSpace=False, orientation="vertical", height=550)
        self.srw_files_box = oasysgui.widgetBox(tab_spdiv, "", addSpace=False, orientation="vertical", height=550)
        self.ascii_box     = oasysgui.widgetBox(tab_spdiv, "", addSpace=False, orientation="vertical", height=550)

    
        ####################################################################################
        # SRW

        tabs_srw = oasysgui.tabWidget(self.srw_box)

        if self.IS_DEVELOP:
            gui.comboBox(self.srw_box, self, "kind_of_sampler", label="Random Generator", labelWidth=250,
                         items=["Simple", "Accurate", "Accurate (SRIO)", ], orientation="horizontal")
        else:
            gui.comboBox(self.srw_box, self, "kind_of_sampler", label="Random Generator", labelWidth=250,
                         items=["Simple", "Accurate"], orientation="horizontal")

        gui.comboBox(self.srw_box, self, "save_srw_result", label="Save SRW results", labelWidth=310,
                     items=["No", "Yes"], orientation="horizontal", callback=self.set_save_file_srw)

        self.save_file_box = oasysgui.widgetBox(self.srw_box, "", addSpace=False, orientation="vertical")
        self.save_file_box_empty = oasysgui.widgetBox(self.srw_box, "", addSpace=False, orientation="vertical", height=55)

        file_box = oasysgui.widgetBox(self.save_file_box, "", addSpace=False, orientation="horizontal", height=25)

        self.le_source_dimension_srw_file = oasysgui.lineEdit(file_box, self, "source_dimension_srw_file", "Source Dimension File", labelWidth=140,  valueType=str, orientation="horizontal")

        gui.button(file_box, self, "...", callback=self.select_source_dimension_file)

        file_box = oasysgui.widgetBox(self.save_file_box, "", addSpace=False, orientation="horizontal", height=25)

        self.le_angular_distribution_srw_file = oasysgui.lineEdit(file_box, self, "angular_distribution_srw_file", "Angular Distribution File", labelWidth=140,  valueType=str, orientation="horizontal")

        gui.button(file_box, self, "...", callback=self.select_angular_distribution_file)

        self.set_save_file_srw()

        tab_ls = oasysgui.createTabPage(tabs_srw, "Undulator Setting")
        tab_wf = oasysgui.createTabPage(tabs_srw, "Wavefront Setting")
        tab_fl = oasysgui.createTabPage(tabs_srw, "Advanced Setting")

        ####################################

        box = oasysgui.widgetBox(tab_fl, "Flux Calculation", addSpace=False, orientation="vertical")

        gui.comboBox(box, self, "use_stokes", label="Integrated Flux", labelWidth=300,
                     items=["From Wavefront", "From Stokes"],
                     sendSelectedValue=False, orientation="horizontal")

        ####################################

        tab_und = oasysgui.tabWidget(tab_ls)

        tab_id   = oasysgui.createTabPage(tab_und, "ID Parameters")
        tab_traj = oasysgui.createTabPage(tab_und, "Trajectory")

        self.tab_pos = oasysgui.tabWidget(tab_id)

        tab_dim   = oasysgui.createTabPage(self.tab_pos, "ID")

        oasysgui.lineEdit(tab_dim, self, "undulator_period", "Period Length [m]", labelWidth=260, valueType=float, orientation="horizontal", callback=self.set_harmonic_energy)
        oasysgui.lineEdit(tab_dim, self, "number_of_periods", "Number of Periods", labelWidth=260, valueType=int, orientation="horizontal")
        oasysgui.lineEdit(tab_dim, self, "horizontal_central_position", "Horizontal Central Position [m]", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(tab_dim, self, "vertical_central_position", "Vertical Central Position [m]", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(tab_dim, self, "longitudinal_central_position", "Longitudinal Central Position [m]", labelWidth=260, valueType=float, orientation="horizontal", callback=self.manage_waist_position)

        self.warning_label = oasysgui.widgetLabel(tab_dim, "  Warning: The source will be positioned at the center\n" +
                                                  "  of the ID: the relative distance of the first optical\n" +
                                                  "  element has to be longitudinally shifted accordingly")
        self.warning_label.setStyleSheet("color: red; font: bold")

        gui.comboBox(tab_dim, self, "magnetic_field_from", label="Magnetic Field", labelWidth=350,
                     items=["From K", "From B"],
                     callback=self.set_magnetic_field,
                     sendSelectedValue=False, orientation="horizontal")

        container = oasysgui.widgetBox(tab_dim, "", addSpace=False, orientation="horizontal")

        horizontal_box = oasysgui.widgetBox(container, "", addSpace=False, orientation="vertical", width=195)
        vertical_box = oasysgui.widgetBox(container,  "", addSpace=False, orientation="vertical", width=155)

        gui.label(horizontal_box, self, "                     Horizontal")
        gui.label(vertical_box, self, "  Vertical")

        self.magnetic_field_box_1_h = oasysgui.widgetBox(horizontal_box, "", addSpace=False, orientation="vertical")
        self.magnetic_field_box_2_h = oasysgui.widgetBox(horizontal_box, "", addSpace=False, orientation="vertical")
        self.magnetic_field_box_1_v = oasysgui.widgetBox(vertical_box, "", addSpace=False, orientation="vertical")
        self.magnetic_field_box_2_v = oasysgui.widgetBox(vertical_box, "", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(self.magnetic_field_box_1_h, self, "Kh", "K", labelWidth=70, valueType=float, orientation="horizontal", callback=self.set_harmonic_energy)
        oasysgui.lineEdit(self.magnetic_field_box_1_v, self, "Kv", " ", labelWidth=2, valueType=float, orientation="horizontal", callback=self.set_harmonic_energy)
        oasysgui.lineEdit(self.magnetic_field_box_2_h, self, "Bh", "B [T]", labelWidth=70, valueType=float, orientation="horizontal", callback=self.set_harmonic_energy)
        oasysgui.lineEdit(self.magnetic_field_box_2_v, self, "Bv", " ", labelWidth=2, valueType=float, orientation="horizontal", callback=self.set_harmonic_energy)

        self.set_magnetic_field()

        oasysgui.lineEdit(horizontal_box, self, "initial_phase_horizontal", "\u03c6\u2080 [rad]", labelWidth=70, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(vertical_box, self, "initial_phase_vertical", " ", labelWidth=2, valueType=float, orientation="horizontal")

        gui.comboBox(horizontal_box, self, "symmetry_vs_longitudinal_position_horizontal", label="Symmetry", labelWidth=70,
                     items=["Symmetrical", "Anti-Symmetrical"],
                     sendSelectedValue=False, orientation="horizontal")

        symmetry_v_box =  oasysgui.widgetBox(vertical_box, "", addSpace=False, orientation="horizontal")
        gui.comboBox(symmetry_v_box, self, "symmetry_vs_longitudinal_position_vertical", label=" ", labelWidth=2,
                     items=["Symmetrical", "Anti-Symmetrical"],
                     sendSelectedValue=False, orientation="horizontal")
        gui.button(symmetry_v_box, self, "?", callback=self.open_help, width=12)

        gui.comboBox(tab_traj, self, "type_of_initialization", label="Trajectory Initialization", labelWidth=140,
                     items=["Automatic", "At Fixed Position", "Sampled from Phase Space"],
                     callback=self.set_type_of_initialization,
                     sendSelectedValue=False, orientation="horizontal")

        self.left_box_3_1 = oasysgui.widgetBox(tab_traj, "", addSpace=False, orientation="vertical", height=160)
        self.left_box_3_2 = oasysgui.widgetBox(tab_traj, "", addSpace=False, orientation="vertical", height=160)

        oasysgui.lineEdit(self.left_box_3_1, self, "moment_x", "x\u2080 [m]", labelWidth=200, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_3_1, self, "moment_y", "y\u2080 [m]", labelWidth=200, valueType=float, orientation="horizontal")

        box = oasysgui.widgetBox(self.left_box_3_1, "", addSpace=False, orientation="horizontal")

        oasysgui.lineEdit(box, self, "moment_z", "z\u2080 [m]", labelWidth=160, valueType=float, orientation="horizontal")
        gui.button(box, self, "Auto", width=35, callback=self.set_z0_default)

        oasysgui.lineEdit(self.left_box_3_1, self, "moment_xp", "x'\u2080 [rad]", labelWidth=200, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_3_1, self, "moment_yp", "y'\u2080 [rad]", labelWidth=200, valueType=float, orientation="horizontal")

        self.set_type_of_initialization()

        left_box_3 = oasysgui.widgetBox(tab_wf, "Divergence Distribution Propagation Parameters", addSpace=False, orientation="vertical")

        box = oasysgui.widgetBox(left_box_3, "", addSpace=False, orientation="horizontal")
        oasysgui.lineEdit(box, self, "source_dimension_wf_h_slit_gap", "H Slit Gap [m]", labelWidth=130, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box, self, "source_dimension_wf_h_slit_c", "  Center [m]", labelWidth=50, valueType=float, orientation="horizontal")
        box = oasysgui.widgetBox(left_box_3, "", addSpace=False, orientation="horizontal")
        oasysgui.lineEdit(box, self, "source_dimension_wf_v_slit_gap", "V Slit Gap [m]", labelWidth=130, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box, self, "source_dimension_wf_v_slit_c", "  Center [m]", labelWidth=50, valueType=float, orientation="horizontal")

        oasysgui.lineEdit(left_box_3, self, "source_dimension_wf_h_slit_points", "H Slit Points", labelWidth=250, valueType=int, orientation="horizontal", callback=self.set_data_x)
        oasysgui.lineEdit(left_box_3, self, "source_dimension_wf_v_slit_points", "V Slit Points", labelWidth=250, valueType=int, orientation="horizontal", callback=self.set_data_y)
        oasysgui.lineEdit(left_box_3, self, "source_dimension_wf_distance", "Propagation Distance [m]\n(relative to the center of the ID)", labelWidth=250, valueType=float, orientation="horizontal")

        left_box_4 = oasysgui.widgetBox(tab_wf, "Size Distribution (Back) Propagation Parameters", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(left_box_4, self, "horizontal_range_modification_factor_at_resizing", "H range modification factor at resizing", labelWidth=290, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(left_box_4, self, "horizontal_resolution_modification_factor_at_resizing", "H resolution modification factor at resizing", labelWidth=290, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(left_box_4, self, "vertical_range_modification_factor_at_resizing", "V range modification factor at resizing", labelWidth=290, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(left_box_4, self, "vertical_resolution_modification_factor_at_resizing", "V resolution modification factor at resizing", labelWidth=290, valueType=float, orientation="horizontal")

        gui.comboBox(tab_wf, self, "auto_expand", label="Auto Expand Slit to Compensate Random Generator", labelWidth=310,
                     items=["No", "Yes"], orientation="horizontal", callback=self.set_auto_expand)

        self.cb_auto_expand_rays = gui.comboBox(tab_wf, self, "auto_expand_rays", label="Auto Increase Number of Rays", labelWidth=310,
                                                items=["No", "Yes"], orientation="horizontal")

        self.set_auto_expand()

        ####################################################################################
        # SRW FILES

        gui.separator(self.srw_files_box)

        file_box = oasysgui.widgetBox(self.srw_files_box, "", addSpace=True, orientation="horizontal", height=45)

        self.le_source_dimension_srw_file = oasysgui.lineEdit(file_box, self, "source_dimension_srw_file", "Source Dimension File", labelWidth=180,  valueType=str, orientation="vertical")

        gui.button(file_box, self, "...", height=45, callback=self.select_source_dimension_file)

        file_box = oasysgui.widgetBox(self.srw_files_box, "", addSpace=True, orientation="horizontal", height=45)

        self.le_angular_distribution_srw_file = oasysgui.lineEdit(file_box, self, "angular_distribution_srw_file", "Angular Distribution File", labelWidth=180,  valueType=str, orientation="vertical")

        gui.button(file_box, self, "...", height=45, callback=self.select_angular_distribution_file)


        ####################################################################################
        # ASCII FILES

        gui.separator(self.ascii_box)

        file_box = oasysgui.widgetBox(self.ascii_box, "", addSpace=True, orientation="horizontal", height=45)

        self.le_x_positions_file = oasysgui.lineEdit(file_box, self, "x_positions_file", "X Positions File", labelWidth=180,  valueType=str, orientation="vertical")

        gui.button(file_box, self, "...", height=45, callback=self.select_x_positions_file)

        file_box = oasysgui.widgetBox(self.ascii_box, "", addSpace=True, orientation="horizontal", height=45)

        self.le_z_positions_file = oasysgui.lineEdit(file_box, self, "z_positions_file", "Z Positions File", labelWidth=180,  valueType=str, orientation="vertical")

        gui.button(file_box, self, "...", height=45, callback=self.select_z_positions_file)

        file_box = oasysgui.widgetBox(self.ascii_box, "", addSpace=True, orientation="horizontal", height=45)

        self.le_x_divergences_file = oasysgui.lineEdit(file_box, self, "x_divergences_file", "X Divergences File", labelWidth=180,  valueType=str, orientation="vertical")

        gui.button(file_box, self, "...", height=45, callback=self.select_x_divergences_file)

        file_box = oasysgui.widgetBox(self.ascii_box, "", addSpace=True, orientation="horizontal", height=45)

        self.le_z_divergences_file = oasysgui.lineEdit(file_box, self, "z_divergences_file", "Z Divergences File", labelWidth=180,  valueType=str, orientation="vertical")

        gui.button(file_box, self, "...", height=45, callback=self.select_z_divergences_file)

        gui.separator(self.ascii_box)

        oasysgui.lineEdit(self.ascii_box, self, "x_positions_factor",   "X Positions UM to Workspace UM", labelWidth=230, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.ascii_box, self, "z_positions_factor",   "Z Positions UM to Workspace UM",  labelWidth=230, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.ascii_box, self, "x_divergences_factor", "X Divergences UM to rad", labelWidth=230, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.ascii_box, self, "z_divergences_factor", "X Divergences UM to rad", labelWidth=230, valueType=float, orientation="horizontal")

        gui.separator(self.ascii_box)

        gui.comboBox(self.ascii_box, self, "combine_strategy", label="2D Distribution Creation Strategy", labelWidth=310,
                     items=["Sqrt(Product)", "Sqrt(Quadratic Sum)", "Convolution", "Average"], orientation="horizontal", callback=self.set_save_file_srw)

        ####################################################################################
        # Utility

        left_box_1 = oasysgui.widgetBox(tab_util, "Auto Setting of Undulator", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(left_box_1, self, "auto_energy", "Set Undulator at Energy [eV]", labelWidth=250, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(left_box_1, self, "auto_harmonic_number", "As Harmonic #",  labelWidth=250, valueType=int, orientation="horizontal")

        button_box = oasysgui.widgetBox(left_box_1, "", addSpace=False, orientation="horizontal")

        gui.button(button_box, self, "Set Kv value", callback=self.auto_set_undulator_V)
        gui.button(button_box, self, "Set Kh value", callback=self.auto_set_undulator_H)
        gui.button(button_box, self, "Set Both K values", callback=self.auto_set_undulator_B)

        gui.rubber(self.controlArea)

        cumulated_plot_tab = oasysgui.createTabPage(self.main_tabs, "Cumulated Plots")

        view_box = oasysgui.widgetBox(cumulated_plot_tab, "Plotting Style", addSpace=False, orientation="horizontal")
        view_box_1 = oasysgui.widgetBox(view_box, "", addSpace=False, orientation="vertical", width=350)

        self.cumulated_view_type_combo = gui.comboBox(view_box_1, self, "cumulated_view_type", label="Show Plots",
                                                      labelWidth=220,
                                                      items=["No", "Yes"],
                                                      callback=self.set_cumulated_plot_quality, sendSelectedValue=False, orientation="horizontal")


        self.cumulated_tabs = oasysgui.tabWidget(cumulated_plot_tab)

        self.initialize_cumulated_tabs()
        self.set_distribution_source()

        gui.rubber(self.mainArea)


    ###################################################################################
    # Listener from HybridCalculator
    ###################################################################################

    def receive_message(self, message: str, data: dict):
        if not message is None:
            print(message)
            self.setStatusMessage(message)

        progress = data.get("progress", None)
        if not progress is None:
            print(f"Progress: {progress}%")
            self.progressBarSet(progress)

    ###################################################################################
    # Utils from HybridCalculator
    ###################################################################################

    def _gamma(self):
        return gamma(electron_beam=self.get_electron_beam())

    def _resonance_energy(self):
        return resonance_energy(electron_beam=self.get_electron_beam(),
                                magnetic_structure=self.get_magnetic_structure(),
                                harmonic=self.harmonic_number)

    def _get_default_initial_z(self):
        return get_default_initial_z(magnetic_structure=self.get_magnetic_structure(), longitudinal_central_position=self.longitudinal_central_position)

    def _is_canted_undulator(self):
        return is_canted_undulator(longitudinal_central_position=self.longitudinal_central_position)

    def _get_source_slit_data(self, direction):
        return get_source_slit_data(HybridUndulatorInputParameters(
                                        auto_expand=self.auto_expand,
                                        source_dimension_wf_h_slit_points = self.source_dimension_wf_h_slit_points,
                                        source_dimension_wf_v_slit_points = self.source_dimension_wf_v_slit_points,
                                        source_dimension_wf_h_slit_gap = self.source_dimension_wf_h_slit_gap,
                                        source_dimension_wf_v_slit_gap = self.source_dimension_wf_v_slit_gap),
                                    direction)

    def _get_which_waist_position_auto(self):
        input_parameters = HybridUndulatorInputParameters(
                                which_waist=self.which_waist,
                                waist_position_auto_h=self.waist_position_auto_h,
                                waist_position_auto_v=self.waist_position_auto_v )
        set_which_waist(input_parameters)

        return input_parameters.waist_position_auto

    #######################################################################################

    def initialize_cumulated_tabs(self):
        current_tab = self.cumulated_tabs.currentIndex()

        self.cumulated_tabs.removeTab(2)
        self.cumulated_tabs.removeTab(1)
        self.cumulated_tabs.removeTab(0)

        self.cumulated_plot_canvas = [None]*3
        self.cumulated_tab = []
        self.cumulated_tab.append(oasysgui.createTabPage(self.cumulated_tabs, "Spectral Flux"))
        self.cumulated_tab.append(oasysgui.createTabPage(self.cumulated_tabs, "Cumulated Power"))
        self.cumulated_tab.append(oasysgui.createTabPage(self.cumulated_tabs, "Power Density"))

        for tab in self.cumulated_tab:
            tab.setFixedHeight(self.IMAGE_HEIGHT)
            tab.setFixedWidth(self.IMAGE_WIDTH)

        self.cumulated_tabs.setCurrentIndex(current_tab)

    def manage_waist_position(self):
        is_canted = self._is_canted_undulator()

        self.warning_label.setVisible(is_canted)
        self.initialize_waist_position_tab(show=is_canted)
        self.initialize_waist_position_plot_tab(show=(is_canted and self.waist_position_calculation == 1))

    def initialize_waist_position_tab(self, show=True):
        if show and self.tab_pos.count() == 1:
            tab_waist   = oasysgui.createTabPage(self.tab_pos, "Waist Position")

            gui.comboBox(tab_waist, self, "waist_position_calculation", label="Waist Position Calculation", labelWidth=310,
                         items=["None", "Automatic", "User Defined"], orientation="horizontal", callback=self.set_waist_position_calculation)

            self.box_none     = oasysgui.widgetBox(tab_waist, "", addSpace=False, orientation="vertical", height=350)
            self.box_auto     = oasysgui.widgetBox(tab_waist, "", addSpace=False, orientation="vertical", height=350)

            gui.comboBox(self.box_auto, self, "waist_back_propagation_parameters", label="Propagation Parameters", labelWidth=250,
                         items=["Same as Source", "Different"], orientation="horizontal", callback=self.set_waist_back_propagation_parameters)

            self.waist_param_box_1 = oasysgui.widgetBox(self.box_auto, "", addSpace=False, orientation="vertical", height=110)
            self.waist_param_box_2 = oasysgui.widgetBox(self.box_auto, "", addSpace=False, orientation="vertical", height=110)

            gui.separator(self.box_auto, height=5)

            oasysgui.lineEdit(self.waist_param_box_2, self, "waist_horizontal_range_modification_factor_at_resizing", "H range modification factor at resizing", labelWidth=290, valueType=float, orientation="horizontal")
            oasysgui.lineEdit(self.waist_param_box_2, self, "waist_horizontal_resolution_modification_factor_at_resizing", "H resolution modification factor at resizing", labelWidth=290, valueType=float, orientation="horizontal")
            oasysgui.lineEdit(self.waist_param_box_2, self, "waist_vertical_range_modification_factor_at_resizing", "V range modification factor at resizing", labelWidth=290, valueType=float, orientation="horizontal")
            oasysgui.lineEdit(self.waist_param_box_2, self, "waist_vertical_resolution_modification_factor_at_resizing", "V resolution modification factor at resizing", labelWidth=290, valueType=float, orientation="horizontal")

            oasysgui.lineEdit(self.box_auto, self, "number_of_waist_fit_points", "Number of Fit Points", labelWidth=290, valueType=int, orientation="horizontal")
            oasysgui.lineEdit(self.box_auto, self, "degree_of_waist_fit", "Degree of Polynomial Fit", labelWidth=290, valueType=int, orientation="horizontal")

            gui.comboBox(self.box_auto, self, "use_sigma_or_fwhm", label="Gaussian size from", labelWidth=250,
                         items=["Sigma", "FWHM"], orientation="horizontal")

            gui.comboBox(self.box_auto, self, "which_waist", label="Use Direction", labelWidth=150,
                         items=["Horizontal", "Vertical", "Both (middle point)"], orientation="horizontal",
                         callback=self.set_which_waist)

            self.set_waist_back_propagation_parameters()

            le = oasysgui.lineEdit(self.box_auto, self, "waist_position_auto", "Waist Position (relative to ID center) [m]", labelWidth=265, valueType=float, orientation="horizontal")
            le.setReadOnly(True)
            font = QFont(le.font())
            font.setBold(True)
            le.setFont(font)
            palette = QPalette(le.palette())
            palette.setColor(QPalette.Text, QColor('dark blue'))
            palette.setColor(QPalette.Base, QColor(243, 240, 160))
            le.setPalette(palette)

            self.box_user_def = oasysgui.widgetBox(tab_waist, "", addSpace=False, orientation="vertical", height=250)

            oasysgui.lineEdit(self.box_user_def, self, "waist_position_user_defined", "Waist Position (relative to ID center) [m]", labelWidth=265, valueType=float, orientation="horizontal")

            self.set_waist_position_calculation()

        elif not show and self.tab_pos.count() == 2:
            self.tab_pos.removeTab(1)

    def initialize_waist_position_plot_tab(self, show=True):
        if show and self.main_tabs.count() == 4:
            waist_tab = oasysgui.createTabPage(self.main_tabs, "Waist Position for Canted Undulator")

            figure = Figure(figsize=(700, 500))

            self.waist_axes = figure.subplots(1, 2)
            self.waist_axes[0].set_title("Horizontal Direction", fontdict={'horizontalalignment': 'right'})
            self.waist_axes[1].set_title("Vertical Direction", fontdict={'horizontalalignment': 'right'})
            self.waist_axes[0].set_xlabel("Position relative to ID center [mm]")
            self.waist_axes[0].set_ylabel("Sigma [um]")
            self.waist_axes[1].set_xlabel("Position relative to ID center [mm]")
            self.waist_axes[1].set_ylabel("Sigma [um]")

            self.waist_figure = FigureCanvas(figure)

            waist_tab.layout().addWidget(self.waist_figure)

        elif not show and self.main_tabs.count() == 5:
            self.main_tabs.removeTab(4)
            self.waist_axes = None

    def set_waist_position_calculation(self):
        self.box_none.setVisible(self.waist_position_calculation==0)
        self.box_auto.setVisible(self.waist_position_calculation==1)
        self.box_user_def.setVisible(self.waist_position_calculation==2)

        self.initialize_waist_position_plot_tab(show=(self.waist_position_calculation == 1))

    def set_waist_back_propagation_parameters(self):
        self.waist_param_box_1.setVisible(self.waist_back_propagation_parameters==0)
        self.waist_param_box_2.setVisible(self.waist_back_propagation_parameters==1)

    def set_cumulated_plot_quality(self):
        if not self.cumulated_power is None:
            self.initialize_cumulated_tabs()

            self._plot_cumulated_results(True)

    def set_auto_expand(self):
        self.cb_auto_expand_rays.setEnabled(self.auto_expand==1)

        self.set_data_xy()

    def set_data_xy(self):
        self.set_data_x()
        self.set_data_y()

    def set_data_x(self):
        source_dimension_wf_h_slit_points, \
        source_dimension_wf_h_slit_gap = self._get_source_slit_data(direction="h")

        x2 = 0.5 * source_dimension_wf_h_slit_gap
        x1 = -x2

        x1 += self.source_dimension_wf_h_slit_c
        x2 += self.source_dimension_wf_h_slit_c

        self.dataX = 1e3 * numpy.linspace(x1, x2, source_dimension_wf_h_slit_points)

    def set_data_y(self):
        source_dimension_wf_v_slit_points, \
        source_dimension_wf_v_slit_gap = self._get_source_slit_data(direction="v")

        y2 = 0.5 * source_dimension_wf_v_slit_gap
        y1 = -y2

        y1 += self.source_dimension_wf_h_slit_c
        y2 += self.source_dimension_wf_h_slit_c

        self.dataY = 1e3*numpy.linspace(y1, y2, source_dimension_wf_v_slit_points)


    ####################################################################################
    # GRAPHICS
    ####################################################################################

    def set_type_of_initialization(self):
        self.left_box_3_1.setVisible(self.type_of_initialization==1)
        self.left_box_3_2.setVisible(self.type_of_initialization!=1)

    def set_z0_default(self):
        self.moment_z = self._get_default_initial_z()

    def auto_set_undulator_V(self):
        self.auto_set_undulator(AutoUndulator.VERTICAL)

    def auto_set_undulator_H(self):
        self.auto_set_undulator(AutoUndulator.HORIZONTAL)

    def auto_set_undulator_B(self):
        self.auto_set_undulator(AutoUndulator.BOTH)

    def auto_set_undulator(self, which=AutoUndulator.VERTICAL):
        if not self.distribution_source == 0: raise Exception("This calculation can be performed only for explicit SRW Calculation")
        congruence.checkStrictlyPositiveNumber(self.auto_energy, "Set Undulator at Energy")
        congruence.checkStrictlyPositiveNumber(self.auto_harmonic_number, "As Harmonic #")
        congruence.checkStrictlyPositiveNumber(self.electron_energy_in_GeV, "Energy")
        congruence.checkStrictlyPositiveNumber(self.undulator_period, "Period Length")

        wavelength = self.auto_harmonic_number*m2ev/self.auto_energy
        K = round(numpy.sqrt(2*(((wavelength*2*self._gamma()**2)/self.undulator_period)-1)), 6)

        if which == AutoUndulator.VERTICAL:
            self.Kv = K
            self.Kh = 0.0

        if which == AutoUndulator.BOTH:
            Kboth = round(K / numpy.sqrt(2), 6)
            self.Kv = Kboth
            self.Kh = Kboth

        if which == AutoUndulator.HORIZONTAL:
            self.Kh = K
            self.Kv = 0.0

        self.set_wf_use_harmonic()

    class ShowHelpDialog(QDialog):

        def __init__(self, parent=None):
            QDialog.__init__(self, parent)
            self.setWindowTitle('Symmetry vs Longitudinal Position')
            layout = QVBoxLayout(self)
            label = QLabel("")

            file = os.path.join(resources.package_dirname("orangecontrib.shadow4_advanced.widgets.tools"), "misc", "symmetry.png")

            label.setPixmap(QPixmap(file))

            bbox = QDialogButtonBox(QDialogButtonBox.Ok)

            bbox.accepted.connect(self.accept)
            layout.addWidget(label)
            layout.addWidget(bbox)

    def open_help(self):
        dialog = HybridUndulator.ShowHelpDialog(parent=self)
        dialog.show()

    def set_magnetic_field(self):
        self.magnetic_field_box_1_h.setVisible(self.magnetic_field_from==0)
        self.magnetic_field_box_2_h.setVisible(self.magnetic_field_from==1)
        self.magnetic_field_box_1_v.setVisible(self.magnetic_field_from==0)
        self.magnetic_field_box_2_v.setVisible(self.magnetic_field_from==1)

        self.set_harmonic_energy()

    def set_harmonic_energy(self):
        if self.distribution_source==0 and self.use_harmonic==0:
            self.harmonic_energy = round(self._resonance_energy(), 2)
        else:
            self.harmonic_energy = numpy.nan

    def set_wf_use_harmonic(self):
        self.use_harmonic_box_1.setVisible(self.use_harmonic==0)
        self.use_harmonic_box_2.setVisible(self.use_harmonic==1)
        self.use_harmonic_box_3.setVisible(self.use_harmonic==2)

        self.set_harmonic_energy()

    def set_distribution_source(self):
        self.srw_box.setVisible(self.distribution_source == 0)
        self.srw_files_box.setVisible(self.distribution_source == 1)
        self.ascii_box.setVisible(self.distribution_source == 2)

        self.set_harmonic_energy()

        if self.distribution_source == 0: self.manage_waist_position()
        else:                             self.initialize_waist_position_plot_tab(show=False)

    def set_save_file_srw(self):
        self.save_file_box.setVisible(self.save_srw_result == 1)
        self.save_file_box_empty.setVisible(self.save_srw_result == 0)

    def select_optimize_file(self):
        self.le_optimize_file_name.setText(oasysgui.selectFileFromDialog(self, self.optimize_file_name, "Open Optimize Source Parameters File"))

    def select_source_dimension_file(self):
        self.le_source_dimension_srw_file.setText(oasysgui.selectFileFromDialog(self, self.source_dimension_srw_file, "Open Source Dimension File"))

    def select_angular_distribution_file(self):
        self.le_angular_distribution_srw_file.setText(oasysgui.selectFileFromDialog(self, self.angular_distribution_srw_file, "Open Angular Distribution File"))

    def select_x_positions_file(self):
        self.le_x_positions_file.setText(oasysgui.selectFileFromDialog(self, self.x_positions_file, "Open X Positions File", file_extension_filter="*.dat, *.txt"))

    def select_z_positions_file(self):
        self.le_z_positions_file.setText(oasysgui.selectFileFromDialog(self, self.z_positions_file, "Open Z Positions File", file_extension_filter="*.dat, *.txt"))

    def select_x_divergences_file(self):
        self.le_x_divergences_file.setText(oasysgui.selectFileFromDialog(self, self.x_divergences_file, "Open X Divergences File", file_extension_filter="*.dat, *.txt"))

    def select_z_divergences_file(self):
        self.le_z_divergences_file.setText(oasysgui.selectFileFromDialog(self, self.z_divergences_file, "Open Z Divergences File", file_extension_filter="*.dat, *.txt"))

    def set_which_waist(self):
        self.waist_position_auto = self._get_which_waist_position_auto()

    ####################################################################################
    # SYNED
    ####################################################################################

    def receive_syned_data(self, data):
        if not data is None:
            try:
                OWSynchrotronSource.receive_syned_data(self, data)

                if data.get_beamline_elements_number() > 0:
                    slit_element = data.get_beamline_element_at(0)
                    slit         = slit_element.get_optical_element()
                    coordinates  = slit_element.get_coordinates()

                    if isinstance(slit, Slit) and isinstance(slit.get_boundary_shape(), Rectangle):
                        rectangle = slit.get_boundary_shape()

                        self.source_dimension_wf_h_slit_gap = numpy.abs(rectangle._x_right - rectangle._x_left)
                        self.source_dimension_wf_h_slit_c = 0.5*(rectangle._x_right + rectangle._x_left)
                        self.source_dimension_wf_v_slit_gap = numpy.abs(rectangle._y_top - rectangle._y_bottom)
                        self.source_dimension_wf_v_slit_c = 0.5 * (rectangle._y_top + rectangle._y_bottom)
                        self.source_dimension_wf_distance = coordinates.p()
            except Exception as exception:
                QMessageBox.critical(self, "Error", str(exception), QMessageBox.Ok)

    def receive_specific_syned_data(self, data): pass

    def populate_fields_from_magnetic_structure(self, magnetic_structure, electron_beam):
        self.Kh                  = magnetic_structure._K_horizontal
        self.Kv                  = magnetic_structure._K_vertical
        self.undulator_period    = magnetic_structure._period_length
        self.number_of_periods   = int(magnetic_structure._number_of_periods)  # SRW needs int
        self.magnetic_field_from = 0

        self.set_magnetic_field()
        self.set_harmonic_energy()

    def check_data(self):
        OWSynchrotronSource.check_data(self)

        if self.use_harmonic == 0:
            if self.distribution_source != 0: raise Exception("Harmonic Energy can be computed only for explicit SRW Calculation")

            self.harmonic_number = congruence.checkStrictlyPositiveNumber(self.harmonic_number, "Harmonic Number")
        elif self.use_harmonic == 2:
            if self.distribution_source != 0: raise Exception("Energy Range can be computed only for explicit SRW Calculation")

            self.energy        = congruence.checkStrictlyPositiveNumber(self.energy, "Photon Energy From")
            self.energy_to     = congruence.checkStrictlyPositiveNumber(self.energy_to, "Photon Energy To")
            self.energy_points = congruence.checkStrictlyPositiveNumber(self.energy_points, "Nr. Energy Values")
            congruence.checkGreaterThan(self.energy_to, self.energy, "Photon Energy To", "Photon Energy From")
        else:
            self.energy = congruence.checkStrictlyPositiveNumber(self.energy, "Photon Energy")

    def check_magnetic_structure(self):
        if self.magnetic_field_from == 0:
            self.Kh = congruence.checkPositiveNumber(self.Kh, "K Horizontal")
            self.Kv = congruence.checkPositiveNumber(self.Kv, "K Vertical")
        else:
            self.Bh = congruence.checkPositiveNumber(self.Bh, "B Horizontal")
            self.Bv = congruence.checkPositiveNumber(self.Bv, "B Vertical")

        self.undulator_period  = congruence.checkStrictlyPositiveNumber(self.undulator_period , "Undulator Period")
        self.number_of_periods = congruence.checkStrictlyPositiveNumber(self.number_of_periods, "Number of Periods")

    def get_magnetic_structure(self):
        return S4HybridUndulator(
            K_vertical=self.Kv if self.magnetic_field_from == 0 else K_from_magnetic_field(self.Bv),
            K_horizontal=self.Kh if self.magnetic_field_from == 0 else K_from_magnetic_field(self.Bh),
            number_of_periods=self.number_of_periods,
            period_length=self.undulator_period
        )

    ####################################################################################
    # LOOPS - TRIGGER
    ####################################################################################

    def set_trigger_parameters_for_sources(self, trigger):
        self.compute_power             = False
        self.energy_step               = None
        self.do_cumulated_calculations = False

        if trigger and trigger.new_object == True:
            if not trigger.has_additional_parameter("start_event"):
                self.cumulated_energies        = None
                self.cumulated_integrated_flux = None
                self.cumulated_power_density   = None
                self.cumulated_power           = None

            if trigger.has_additional_parameter("energy_value") and trigger.has_additional_parameter("energy_step"):
                self.compute_power             = True
                self.use_harmonic              = 1
                self.distribution_source       = 0
                self.save_srw_result           = 0
                self.do_cumulated_calculations = True

                if trigger.has_additional_parameter("start_event") and trigger.get_additional_parameter("start_event") == True:
                    self.cumulated_energies        = None
                    self.cumulated_integrated_flux = None
                    self.cumulated_power_density   = None
                    self.cumulated_power           = None

                self.energy       = trigger.get_additional_parameter("energy_value")
                self.energy_step  = trigger.get_additional_parameter("energy_step")
                self.power_step   = trigger.get_additional_parameter("power_step")
                self.current_step = trigger.get_additional_parameter("current_step")
                self.total_steps  = trigger.get_additional_parameter("total_steps")
                self.start_event  = trigger.get_additional_parameter("start_event")

                self.set_wf_use_harmonic()
                self.set_distribution_source()
                self.set_save_file_srw()

        OWSynchrotronSource.set_trigger_parameters_for_sources(self, trigger)

    ####################################################################################
    # SHADOW
    ####################################################################################

    def run_shadow4(self, scanning_data = None):
        if not scanning_data: scanning_data = None

        self.setStatusMessage("")
        self.progressBarInit()

        sys.stdout = EmittingStream(textWritten=self._write_stdout)

        try:
            self.check_data()

            hybrid_input_parameters = HybridUndulatorInputParameters(
                electron_beam                                               = self.get_electron_beam(),
                magnetic_structure                                          = self.get_magnetic_structure(),
                number_of_rays                                              = self.number_of_rays,
                seed                                                        = self.seed,
                use_harmonic                                                = self.use_harmonic,
                harmonic_number                                             = self.harmonic_number,
                energy                                                      = self.energy,
                energy_to                                                   = self.energy_to,
                energy_points                                               = self.energy_points,
                initial_phase_vertical                                      = self.initial_phase_vertical,
                initial_phase_horizontal                                    = self.initial_phase_horizontal,
                symmetry_vs_longitudinal_position_vertical                  = self.symmetry_vs_longitudinal_position_vertical,
                symmetry_vs_longitudinal_position_horizontal                = self.symmetry_vs_longitudinal_position_horizontal,
                horizontal_central_position                                 = self.horizontal_central_position,
                vertical_central_position                                   = self.vertical_central_position,
                longitudinal_central_position                               = self.longitudinal_central_position,
                type_of_initialization                                      = self.type_of_initialization,
                use_stokes                                                  = self.use_stokes,
                auto_expand                                                 = self.auto_expand,
                auto_expand_rays                                            = self.auto_expand_rays,
                source_dimension_wf_h_slit_gap                              = self.source_dimension_wf_h_slit_gap,
                source_dimension_wf_v_slit_gap                              = self.source_dimension_wf_v_slit_gap,
                source_dimension_wf_h_slit_c                                = self.source_dimension_wf_h_slit_c,
                source_dimension_wf_v_slit_c                                = self.source_dimension_wf_v_slit_c,
                source_dimension_wf_h_slit_points                           = self.source_dimension_wf_h_slit_points,
                source_dimension_wf_v_slit_points                           = self.source_dimension_wf_v_slit_points,
                source_dimension_wf_distance                                = self.source_dimension_wf_distance,
                horizontal_range_modification_factor_at_resizing            = self.horizontal_range_modification_factor_at_resizing,
                horizontal_resolution_modification_factor_at_resizing       = self.horizontal_resolution_modification_factor_at_resizing,
                vertical_range_modification_factor_at_resizing              = self.vertical_range_modification_factor_at_resizing,
                vertical_resolution_modification_factor_at_resizing         = self.vertical_resolution_modification_factor_at_resizing,
                waist_position_calculation                                  = self.waist_position_calculation,
                waist_position                                              = self.waist_position,
                waist_position_auto                                         = self.waist_position_auto,
                waist_position_auto_h                                       = self.waist_position_auto_h,
                waist_position_auto_v                                       = self.waist_position_auto_v,
                waist_back_propagation_parameters                           = self.waist_back_propagation_parameters,
                waist_horizontal_range_modification_factor_at_resizing      = self.waist_horizontal_range_modification_factor_at_resizing,
                waist_horizontal_resolution_modification_factor_at_resizing = self.waist_horizontal_resolution_modification_factor_at_resizing,
                waist_vertical_range_modification_factor_at_resizing        = self.waist_vertical_range_modification_factor_at_resizing,
                waist_vertical_resolution_modification_factor_at_resizing   = self.waist_vertical_resolution_modification_factor_at_resizing,
                which_waist                                                 = self.which_waist,
                number_of_waist_fit_points                                  = self.number_of_waist_fit_points,
                degree_of_waist_fit                                         = self.degree_of_waist_fit,
                use_sigma_or_fwhm                                           = self.use_sigma_or_fwhm,
                waist_position_user_defined                                 = self.waist_position_user_defined,
                kind_of_sampler                                             = self.kind_of_sampler,
                save_srw_result                                             = self.save_srw_result,
                source_dimension_srw_file                                   = self.source_dimension_srw_file,
                angular_distribution_srw_file                               = self.angular_distribution_srw_file,
                x_positions_file                                            = self.x_positions_file,
                z_positions_file                                            = self.z_positions_file,
                x_positions_factor                                          = self.x_positions_factor,
                z_positions_factor                                          = self.z_positions_factor,
                x_divergences_file                                          = self.x_divergences_file,
                z_divergences_file                                          = self.z_divergences_file,
                x_divergences_factor                                        = self.x_divergences_factor,
                z_divergences_factor                                        = self.z_divergences_factor,
                combine_strategy                                            = self.combine_strategy,
                distribution_source                                         = self.distribution_source,
                energy_step                                                 = self.energy_step,
                power_step                                                  = self.power_step,
                compute_power                                               = self.compute_power,
                integrated_flux                                             = self.integrated_flux,
                power_density                                               = self.power_density)

            light_source = S4HybridUndulatorLightSource(name="Shadow4/SRW Undulator Source",
                                                        hybrid_input_parameters=hybrid_input_parameters,
                                                        calculation_listener=self)

            # script
            script = light_source.to_python_code()
            script += "\n\n# test plot\nfrom srxraylib.plot.gol import plot_scatter"
            script += "\nrays = beam.get_rays()"
            script += "\nplot_scatter(1e6 * rays[:, 0], 1e6 * rays[:, 2], title='(X,Z) in microns')"
            self.shadow4_script.set_code(script)

            params = {"do_cumulated_calculations" : self.do_cumulated_calculations}
            output_beam              = light_source.get_beam(**params)
            hybrid_output_parameters = light_source.get_output_parameters()

            self.cumulated_energies        = hybrid_output_parameters.cumulated_energies
            self.cumulated_integrated_flux = hybrid_output_parameters.cumulated_integrated_flux
            self.cumulated_power_density   = hybrid_output_parameters.cumulated_power_density
            self.cumulated_power           = hybrid_output_parameters.cumulated_power

            self.moment_x        = hybrid_output_parameters.moment_x
            self.moment_y        = hybrid_output_parameters.moment_y
            self.moment_z        = hybrid_output_parameters.moment_z
            self.moment_xp       = hybrid_output_parameters.moment_xp
            self.moment_yp       = hybrid_output_parameters.moment_yp

            #
            # send beam and trigger
            #
            output_data = ShadowData(beam=output_beam,
                                     number_of_rays=self.number_of_rays,
                                     beamline=S4Beamline(light_source=light_source))

            self.setStatusMessage("Plotting Results")

            self.progressBarSet(80)

            self._plot_results(output_beam, None, progressBarValue=80)

            if self.distribution_source == 0 and \
                    is_canted_undulator(hybrid_input_parameters) and \
                    self.waist_position_calculation == 1:

                    self.waist_position_auto_h = hybrid_input_parameters.waist_position_auto_h
                    self.waist_position_auto_v = hybrid_input_parameters.waist_position_auto_v
                    self.waist_position_auto   = hybrid_input_parameters.waist_position_auto
                    self.waist_position        = hybrid_input_parameters.waist_position

                    self._plot_waist(hybrid_output_parameters)

            self._plot_cumulated_results(self.do_cumulated_calculations)

            self.setStatusMessage("")

            if self.compute_power and self.energy_step and hybrid_output_parameters.total_power:
                additional_parameters = {}
                additional_parameters["total_power"]        = hybrid_output_parameters.total_power
                additional_parameters["photon_energy_step"] = self.energy_step
                additional_parameters["current_step"]       = self.current_step
                additional_parameters["total_steps"]        = self.total_steps

                output_data.scanning_data = ShadowData.ScanningData("photon_energy", self.energy, "Energy for Power Calculation", "eV", additional_parameters)

            self.Outputs.shadow_data.send(output_data)
            self.Outputs.trigger.send(TriggerIn(new_object=True))
        except Exception as exception:
            QMessageBox.critical(self, "Error", str(exception), QMessageBox.Ok)

            if self.IS_DEVELOP: raise exception

        self.progressBarFinished()

    def _cumulated_plot_data1D(self, dataX, dataY, plot_canvas_index, title="", xtitle="", ytitle=""):
        if self.cumulated_plot_canvas[plot_canvas_index] is None:
            self.cumulated_plot_canvas[plot_canvas_index] = oasysgui.plotWindow()
            self.cumulated_tab[plot_canvas_index].layout().addWidget(self.cumulated_plot_canvas[plot_canvas_index])

        self.cumulated_plot_canvas[plot_canvas_index].addCurve(dataX, dataY,)

        self.cumulated_plot_canvas[plot_canvas_index].resetZoom()
        self.cumulated_plot_canvas[plot_canvas_index].setXAxisAutoScale(True)
        self.cumulated_plot_canvas[plot_canvas_index].setYAxisAutoScale(True)
        self.cumulated_plot_canvas[plot_canvas_index].setGraphGrid(False)

        self.cumulated_plot_canvas[plot_canvas_index].setXAxisLogarithmic(False)
        self.cumulated_plot_canvas[plot_canvas_index].setYAxisLogarithmic(False)
        self.cumulated_plot_canvas[plot_canvas_index].setGraphXLabel(xtitle)
        self.cumulated_plot_canvas[plot_canvas_index].setGraphYLabel(ytitle)
        self.cumulated_plot_canvas[plot_canvas_index].setGraphTitle(title)

    def _cumulated_plot_data2D(self, data2D, dataX, dataY, plot_canvas_index, title="", xtitle="", ytitle=""):

        if self.cumulated_plot_canvas[plot_canvas_index] is None:
            self.cumulated_plot_canvas[plot_canvas_index] = Plot2D()
            self.cumulated_tab[plot_canvas_index].layout().addWidget(self.cumulated_plot_canvas[plot_canvas_index])

        origin = (dataX[0],dataY[0])
        scale = (dataX[1]-dataX[0], dataY[1]-dataY[0])

        data_to_plot = data2D.T

        colormap = {"name":"temperature", "normalization":"linear", "autoscale":True, "vmin":0, "vmax":0, "colors":256}

        self.cumulated_plot_canvas[plot_canvas_index].resetZoom()
        self.cumulated_plot_canvas[plot_canvas_index].setXAxisAutoScale(True)
        self.cumulated_plot_canvas[plot_canvas_index].setYAxisAutoScale(True)
        self.cumulated_plot_canvas[plot_canvas_index].setGraphGrid(False)
        self.cumulated_plot_canvas[plot_canvas_index].setKeepDataAspectRatio(True)
        self.cumulated_plot_canvas[plot_canvas_index].yAxisInvertedAction.setVisible(False)

        self.cumulated_plot_canvas[plot_canvas_index].setXAxisLogarithmic(False)
        self.cumulated_plot_canvas[plot_canvas_index].setYAxisLogarithmic(False)
        self.cumulated_plot_canvas[plot_canvas_index].getMaskAction().setVisible(False)
        self.cumulated_plot_canvas[plot_canvas_index].getRoiAction().setVisible(False)
        self.cumulated_plot_canvas[plot_canvas_index].getColormapAction().setVisible(False)
        self.cumulated_plot_canvas[plot_canvas_index].setKeepDataAspectRatio(False)



        self.cumulated_plot_canvas[plot_canvas_index].addImage(numpy.array(data_to_plot),
                                                     legend="active",
                                                     scale=scale,
                                                     origin=origin,
                                                     colormap=colormap,
                                                     replace=True)

        self.cumulated_plot_canvas[plot_canvas_index].setActiveImage("active")
        self.cumulated_plot_canvas[plot_canvas_index].setGraphXLabel(xtitle)
        self.cumulated_plot_canvas[plot_canvas_index].setGraphYLabel(ytitle)
        self.cumulated_plot_canvas[plot_canvas_index].setGraphTitle(title)

    def _plot_cumulated_results(self, do_cumulated_calculations):
        if not self.cumulated_view_type == 0 and do_cumulated_calculations==True:
            try:
                self.cumulated_view_type_combo.setEnabled(False)

                total_power = str(round(self.cumulated_power[-1], 2))

                self._cumulated_plot_data1D(self.cumulated_energies, self.cumulated_integrated_flux, 0, "Spectral Flux", "Energy [eV]", "Flux [ph/s/0.1%BW]")
                self._cumulated_plot_data1D(self.cumulated_energies, self.cumulated_power, 1,
                                           "Cumulated Power (Total = " + total_power + " W)", "Energy [eV]", "Power [W]")
                self._cumulated_plot_data2D(self.cumulated_power_density, self.dataX, self.dataY, 2,
                                           "Power Density [W/mm^2] (Total Power = " + total_power + " W)", "X [mm]", "Y [mm]")

                self.cumulated_view_type_combo.setEnabled(True)
            except Exception as e:
                self.cumulated_view_type_combo.setEnabled(True)

                raise Exception("Data not plottable: exception: " + str(e))


    def _plot_waist(self, hybrid_output_parameters: HybridUndulatorOutputParameters):
        def plot(direction, positions, sizes_e, sizes_ph, size_ph_an, sizes_tot, waist_position, waist_size):
            self.waist_axes[direction].clear()
            self.waist_axes[direction].set_title(("Horizontal" if direction == 0 else "Vertical") + " Direction\n" +
                                                   "Source size: " + str(round(waist_size * 1e6, 2)) + " " + r'$\mu$' + "m \n" +
                                                   "at " + str(round(waist_position * 1e3, 1)) + " mm from the ID center")

            self.waist_axes[direction].plot(positions * 1e3, sizes_e * 1e6, label='electron', color='g')
            self.waist_axes[direction].plot(positions * 1e3, sizes_ph * 1e6, label='photon', color='b')
            self.waist_axes[direction].plot(positions * 1e3, size_ph_an * 1e6, '--', label='photon (analytical)', color='b')
            self.waist_axes[direction].plot(positions * 1e3, sizes_tot * 1e6, label='total', color='r')
            self.waist_axes[direction].plot([waist_position * 1e3], [waist_size * 1e6], 'bo', label="waist")
            self.waist_axes[direction].set_xlabel("Position relative to ID center [mm]")
            self.waist_axes[direction].set_ylabel("Sigma [um]")
            self.waist_axes[direction].legend()

        plot(0,
             hybrid_output_parameters.positions,
             hybrid_output_parameters.sizes_e_x,
             hybrid_output_parameters.sizes_ph_x,
             hybrid_output_parameters.sizes_ph_an_x,
             hybrid_output_parameters.sizes_tot_x,
             hybrid_output_parameters.waist_position_x,
             hybrid_output_parameters.waist_size_x)
        plot(1,
             hybrid_output_parameters.positions,
             hybrid_output_parameters.sizes_e_y,
             hybrid_output_parameters.sizes_ph_y,
             hybrid_output_parameters.sizes_ph_an_y,
             hybrid_output_parameters.sizes_tot_y,
             hybrid_output_parameters.waist_position_y,
             hybrid_output_parameters.waist_size_y)

        try: self.waist_figure.draw()
        except Exception as e: print(e)

add_widget_parameters_to_module(__name__)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# #########################################################################
# Copyright (c) 2020, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2020. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# #########################################################################

import sys, numpy, os

from AnyQt.QtCore import Qt
from AnyQt.QtWidgets import QMessageBox, QLabel, QSizePolicy
from AnyQt.QtGui import QPixmap
from matplotlib.backend_bases import button_press_handler

from matplotlib.patches import FancyArrowPatch, ArrowStyle

import orangecanvas.resources as resources

from orangewidget import gui
from orangewidget.widget import Input, Output
from orangewidget.settings import Setting

from oasys2.canvas.util.canvas_util import add_widget_parameters_to_module
from oasys2.widget.widget import OWAction
from oasys2.widget import gui as oasysgui
from oasys2.widget.util.exchange import DataExchangeObject

from srxraylib.util.histograms import get_fwhm

from shadow4.beam.s4_beam import S4Beam

try:
    from orangecontrib.shadow4.util.shadow4_objects import ShadowData
    from orangecontrib.shadow4.util.shadow4_util import ShadowCongruence, ShadowPlot
    from orangecontrib.shadow4.widgets.gui.ow_automatic_element import AutomaticElement
except ImportError:
    pass

TO_MM = 1e3

class FluxCalculator(AutomaticElement):

    name = "Flux Calculator"
    description = "Tools: Flux Calculator"
    icon = "icons/flux.png"
    maintainer = "Luca Rebuffi"
    maintainer_email = "lrebuffi(@at@)anl.gov"
    priority = 8
    category = "User Defined"
    keywords = ["data", "file", "load", "read"]

    class Inputs:
        shadow_data = Input("Shadow Data", ShadowData, default=True, auto_summary=False)
        energy_spectrum = Input("Energy Spectrum", DataExchangeObject, default=True, auto_summary=False)

    class Outputs:
        shadow_data = Output("Shadow Data", ShadowData, default=True, auto_summary=False)

    want_main_area = 0
    want_control_area = 1

    input_data     = None
    input_spectrum = None
    flux_index = -1

    bandwidth_calculation = Setting(0)

    e_min = Setting(0.0)
    e_max = Setting(0.0)
    n_bins = Setting(200)

    usage_path = os.path.join(resources.package_dirname("orangecontrib.shadow4_advanced.widgets.tools"), "misc", "flux_calculator.png")

    def __init__(self):
        super(FluxCalculator, self).__init__()

        self.runaction = OWAction("Calculate Flux", self)
        self.runaction.triggered.connect(self.calculate_flux)
        self.addAction(self.runaction)

        self.setMaximumWidth(self.CONTROL_AREA_WIDTH+10)
        self.setMaximumHeight(660)

        button_box = oasysgui.widgetBox(self.controlArea, "", orientation="horizontal", width=self.CONTROL_AREA_WIDTH-5)
        gui.button(button_box, self, "Calculate Flux", callback=self.calculate_flux, height=45)

        tabs_setting = oasysgui.tabWidget(self.controlArea)
        tabs_setting.setFixedHeight(510)
        tabs_setting.setFixedWidth(self.CONTROL_AREA_WIDTH-8)

        tab_ban = oasysgui.createTabPage(tabs_setting, "Beam Bandwidth")
        tab_out = oasysgui.createTabPage(tabs_setting, "Flux Calculation Results")
        tab_usa = oasysgui.createTabPage(tabs_setting, "Use of the Widget")
        tab_usa.setStyleSheet("background-color: white;")

        bandwidth_box = oasysgui.widgetBox(tab_ban, "Bandwidth", addSpace=True, orientation="vertical")

        gui.comboBox(bandwidth_box, self, "bandwidth_calculation", label="BW Calculation Mode", labelWidth=260,
                     items=["Automatic", "Manual"], sendSelectedValue=False, orientation="horizontal", callback=self.set_bw_calculation_mode)

        self.bandwidth_box_1 = oasysgui.widgetBox(bandwidth_box, "", addSpace=True, orientation="vertical", height=90)
        self.bandwidth_box_2 = oasysgui.widgetBox(bandwidth_box, "", addSpace=True, orientation="vertical", height=90)

        oasysgui.lineEdit(self.bandwidth_box_1, self, "e_min", "Energy min", labelWidth=200, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.bandwidth_box_1, self, "e_max", "Energy max", labelWidth=200, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.bandwidth_box_1, self, "n_bins", "Number of Bins", labelWidth=200, valueType=int, orientation="horizontal")

        self.histo_energy = oasysgui.plotWindow(resetzoom=False,
                                                autoScale=False,
                                                logScale=False,
                                                grid=False,
                                                curveStyle=False,
                                                colormap=False,
                                                aspectRatio=False,
                                                yInverted=False,
                                                copy=False,
                                                save=False,
                                                print_=False,
                                                control=False,
                                                position=False,
                                                roi=False,
                                                mask=False,
                                                fit=False)
        self.histo_energy.setDefaultPlotLines(True)
        self.histo_energy._toolbar.setVisible(False)
        self.histo_energy._interactiveModeToolBar.setVisible(False)
        self.histo_energy._outputToolBar.setVisible(False)
        self.histo_energy.group.setVisible(False)
        self.histo_energy._colorbar.setVisible(False)
        self.histo_energy.setActiveCurveStyle(color='blue')
        self.histo_energy.setMinimumWidth(380)

        tab_ban.layout().addWidget(self.histo_energy)

        self.set_bw_calculation_mode()

        self.text = oasysgui.textArea(width=self.CONTROL_AREA_WIDTH-22, height=470)

        tab_out.layout().addWidget(self.text)

        usage_box = oasysgui.widgetBox(tab_usa, "", addSpace=True, orientation="horizontal")

        label = QLabel("")
        label.setAlignment(Qt.AlignCenter)
        label.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        label.setPixmap(QPixmap(self.usage_path))

        usage_box.layout().addWidget(label)

        gui.rubber(self.controlArea)

    def set_bw_calculation_mode(self):
        self.bandwidth_box_1.setVisible(self.bandwidth_calculation==1)
        self.bandwidth_box_2.setVisible(self.bandwidth_calculation==0)

    @Inputs.shadow_data
    def set_shadow_data(self, shadow_data):
        try:
            if ShadowCongruence.check_empty_data(shadow_data):
                if ShadowCongruence.check_good_beam(shadow_data.beam):
                    self.input_data = shadow_data

                    if self.is_automatic_run: self.calculate_flux()
        except Exception as exception:
            QMessageBox.critical(self, "Error", str(exception), QMessageBox.Ok)

            if self.IS_DEVELOP: raise exception

    @Inputs.energy_spectrum
    def set_energy_spectrum(self, data):
        if not data is None:
            try:
                if data.get_program_name() == "XOPPY":
                    if data.get_widget_name() == "UNDULATOR_FLUX" or data.get_widget_name() == "XWIGGLER" or data.get_widget_name() == "WS":
                        self.flux_index = 1
                    elif data.get_widget_name() == "BM":
                        self.flux_index = 5
                    else:
                        raise Exception("Connect to one of the following XOPPY widgets: Undulator Spectrum, BM, XWIGGLER, WS")

                    self.input_spectrum = data.get_content('xoppy_data')
                elif data.get_program_name() == "SRW":
                    if data.get_widget_name() == "UNDULATOR_SPECTRUM":
                        self.flux_index = 1
                    else:
                        raise Exception("Connect to one of the following SRW widgets: Undulator Spectrum")

                    self.input_spectrum = data.get_content('srw_data')
                else:
                    raise ValueError("Widget accept data from the following Add-ons: XOPPY, SRW")

                if self.is_automatic_run: self.calculate_flux()
            except Exception as exception:
                QMessageBox.critical(self, "Error", str(exception), QMessageBox.Ok)

                if self.IS_DEVELOP: raise exception

    def calculate_flux(self):
        if not self.input_data is None and not self.input_spectrum is None:
            try:
                if self.bandwidth_calculation==1:
                    if self.e_min >= self.e_max: raise ValueError("Energy min should be < Energy max")
                    if self.n_bins <= 0 : raise ValueError("Nr. bins should be > 0")

                    erange = [self.e_min, self.e_max]
                    nbins  = self.n_bins
                else:
                    erange = None
                    nbins  = 200

                shadow_beam = self.input_data.beam

                self.plot_histo(shadow_beam=shadow_beam, erange=erange, nbins=nbins)
                flux_factor, resolving_power, energy, ttext = calculate_flux_factor_and_resolving_power(shadow_beam=shadow_beam, erange=erange, nbins=nbins)

                total_text = ttext

                flux_at_sample, ttext = calculate_flux_at_sample(self.input_spectrum, self.flux_index, flux_factor, energy)

                ticket = shadow_beam.histo2(1, 3, nbins=100, nolost=1, ref=23)

                dx = ticket['fwhm_v'] * TO_MM
                dy = ticket['fwhm_h'] * TO_MM

                total_text += "\n" + ttext

                total_text += "\n\n ---> Integrated Flux : %g"%flux_at_sample + " ph/s"
                total_text += "\n ---> <Flux Density>  : %g"%(flux_at_sample/(dx*dy)) + " ph/s/mm^2"
                total_text += "\n ---> Resolving Power : %g"%resolving_power

                self.text.clear()
                self.text.setText(total_text)

                self.Outputs.shadow_data.send(self.input_data)
            except Exception as exception:
                QMessageBox.critical(self, "Error", str(exception), QMessageBox.Ok)

                if self.IS_DEVELOP: raise exception

    def plot_histo(self, shadow_beam, erange=None, nbins=200):
        self.histo_energy.clear()

        ticket = shadow_beam.histo1(26, nbins=nbins, xrange=erange, nolost=1, ref=23)

        ticket['fwhm'], ticket['fwhm_quote'], ticket['fwhm_coordinates'] = get_fwhm(ticket['histogram'], ticket['bin_center'])

        histogram = ticket['histogram_path']
        bins = ticket['bin_path']
        if ticket['fwhm'] == None: ticket['fwhm'] = 0.0

        self.histo_energy.addCurve(bins, histogram, "Energy", symbol='', color='blue', replace=True) #'+', '^', ','
        self.histo_energy.setGraphXLabel("Energy")
        self.histo_energy.setGraphYLabel("Intensity")
        self.histo_energy.setGraphTitle("Bandwidth: " + str(round(ticket['fwhm'], 4)) + " eV")
        self.histo_energy.setInteractiveMode(mode='zoom')

        n_patches = len(self.histo_energy._backend.ax.patches)
        if (n_patches > 0): self.histo_energy._backend.ax.patches[n_patches - 1].remove()

        if not ticket['fwhm'] == 0.0:
            x_fwhm_i, x_fwhm_f = ticket['fwhm_coordinates']
            x_fwhm_i, x_fwhm_f = x_fwhm_i, x_fwhm_f
            y_fwhm             = ticket['fwhm_quote']


            self.histo_energy._backend.ax.add_patch(FancyArrowPatch([x_fwhm_i, y_fwhm],
                                                                    [x_fwhm_f, y_fwhm],
                                                                    arrowstyle=ArrowStyle.CurveAB(head_width=2, head_length=4),
                                                                    color='b',
                                                                    linewidth=1.5))
        if min(histogram) < 0: self.histo_energy.setGraphYLimits(min(histogram), max(histogram))
        else:                  self.histo_energy.setGraphYLimits(0, max(histogram))

        self.histo_energy.replot()


def calculate_flux_factor_and_resolving_power(shadow_beam, nbins=200, erange=None):
    ticket = shadow_beam.histo1(26, nbins=nbins, xrange=None, nolost=1)

    energy_min = ticket['xrange'][0]
    energy_max = ticket['xrange'][-1]

    Denergy_source = numpy.abs(energy_max - energy_min)
    energy = numpy.average([energy_min, energy_max])

    if Denergy_source == 0.0:
        raise ValueError("This calculation is not possibile for a single energy value")

    ticket = shadow_beam.histo1(26, nbins=nbins, nolost=1, xrange=erange, ref=23)

    initial_intensity = shadow_beam.get_number_of_rays()
    final_intensity   = ticket['intensity']
    efficiency = final_intensity/initial_intensity
    bandwidth  = ticket['fwhm']

    if bandwidth == 0.0 or bandwidth is None: raise ValueError("Bandwidth is 0.0 or None: calculation not possible")

    resolving_power = energy/bandwidth

    if Denergy_source < 4*bandwidth:
        raise ValueError("Source \u0394E (" + str(round(Denergy_source, 2)) + " eV) should be at least 4 times bigger than the bandwidth (" + str(round(bandwidth, 3)) + " eV)")

    text = "\n# SOURCE ---------\n"
    text += "\n Source Central Energy: %g"%round(energy, 2) + " eV"
    text += "\n Source Energy Range  : %g - %g"%(round(energy_min, 2), round(energy_max, 2)) + " eV"
    text += "\n Source \u0394E: %g"%round(Denergy_source, 2) + " eV"

    text += "\n\n# BEAMLINE ---------\n"
    text += "\n Shadow Intensity (Initial): %g"%initial_intensity
    text += "\n Shadow Intensity (Final)  : %g"%final_intensity
    text += "\n"
    text += "\n Efficiency: %g"%round(100*efficiency, 3) + "%"
    text += "\n Bandwidth (at the Image Plane): %g"%round(bandwidth, 3) + " eV"

    beamline_bandwidth = Denergy_source * efficiency

    flux_factor = beamline_bandwidth / (1e-3*energy)

    return flux_factor, resolving_power, energy, text

def calculate_flux_at_sample(spectrum, flux_index, flux_factor, energy):
    if energy < spectrum[0, 0] or energy > spectrum[-1, 0]: raise ValueError("Spectrum does not contained central energy")
    interpolated_flux = numpy.interp([energy],
                                     spectrum[:, 0],
                                     spectrum[:, flux_index],
                                     left=spectrum[:, flux_index][0],
                                     right=spectrum[:, flux_index][-1])[0]

    text = "\n# FLUX INTERPOLATION ---------\n"
    text += "\n Initial Flux from Source: %g"%interpolated_flux + " ph/s/0.1%bw"

    return interpolated_flux*flux_factor, text


add_widget_parameters_to_module(__name__)





#!/usr/bin/env python
# -*- coding: utf-8 -*-
# #########################################################################
# Copyright (c) 2020, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2020. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# #########################################################################

from orangewidget import gui
from orangewidget.settings import Setting

from oasys2.canvas.util.canvas_util import add_widget_parameters_to_module

from oasys2.widget import gui as oasysgui
from oasys2.widget.util import congruence
from oasys2.widget.gui import ConfirmDialog

try:
    from orangecontrib.shadow4.util.shadow4_util import ShadowCongruence, ShadowPlot
except ImportError:
    pass

import scipy.constants as codata

from orangecontrib.shadow4_advanced.widgets.gui.power_plot_xy import AbstractPowerPlotXY, PowerPlotXYWidget

class PowerPlotXY(AbstractPowerPlotXY):
    name = "Power Density - Undulator"
    description = "Display Data Tools: Power Plot XY - Undulator"
    icon = "icons/plot_xy_power.png"
    priority = 6.3

    keep_result=Setting(1)
    autosave_partial_results = Setting(0)

    autosave = Setting(0)
    autosave_file_name = Setting("autosave_power_density.hdf5")

    autosave_file = None

    def __init__(self):
        super().__init__()

    def _set_additional_boxes(self, tab_gen):
        autosave_box = oasysgui.widgetBox(tab_gen, "Autosave", addSpace=True, orientation="vertical", height=85)

        gui.comboBox(autosave_box, self, "autosave", label="Save automatically plot into file", labelWidth=250,
                                         items=["No", "Yes"],
                                         sendSelectedValue=False, orientation="horizontal", callback=self.set_autosave)

        self.autosave_box_1 = oasysgui.widgetBox(autosave_box, "", addSpace=False, orientation="horizontal", height=25)
        self.autosave_box_2 = oasysgui.widgetBox(autosave_box, "", addSpace=False, orientation="horizontal", height=25)

        self.le_autosave_file_name = oasysgui.lineEdit(self.autosave_box_1, self, "autosave_file_name", "File Name", labelWidth=100,  valueType=str, orientation="horizontal")

        gui.button(self.autosave_box_1, self, "...", callback=self.select_autosave_file)

        incremental_box = oasysgui.widgetBox(tab_gen, "Incremental Result", addSpace=True, orientation="vertical", height=120)

        gui.comboBox(incremental_box, self, "keep_result", label="Keep Result", labelWidth=250,
                     items=["No", "Yes"], sendSelectedValue=False, orientation="horizontal", callback=self.set_autosave)

        self.cb_autosave_partial_results = gui.comboBox(incremental_box, self, "autosave_partial_results", label="Save partial plots into file", labelWidth=250,
                                                        items=["No", "Yes"], sendSelectedValue=False, orientation="horizontal")

        gui.button(incremental_box, self, "Clear", callback=self.clear_results)

        self.set_autosave()

    def clear_results(self, interactive=True):
        if not interactive: proceed = True
        else: proceed = ConfirmDialog.confirmed(parent=self)

        if proceed:
            self.input_data = None
            self.cumulated_ticket = None
            self.plotted_ticket = None
            self.energy_min = None
            self.energy_max = None
            self.energy_step = None
            self.total_power = None
            self.cumulated_total_power = None

            if not self.autosave_file is None:
                self.autosave_file.close()
                self.autosave_file = None

            if not self.plot_canvas is None:
                self.plot_canvas.clear()

    def set_autosave(self):
        self.autosave_box_1.setVisible(self.autosave==1)
        self.autosave_box_2.setVisible(self.autosave==0)

        self.cb_autosave_partial_results.setEnabled(self.autosave==1 and self.keep_result==1)

    def select_autosave_file(self):
        file_name = oasysgui.selectSaveFileFromDialog(self, "Select File", default_file_name="", file_extension_filter="HDF5 Files (*.hdf5 *.h5 *.hdf)")
        self.le_autosave_file_name.setText("" if file_name is None else file_name)

    #########################################################
    # I/O

    def _analyze_input_data(self, input_data):
        if not input_data.scanning_data is None and input_data.scanning_data.has_additional_parameter("total_power"):
            self.input_data = input_data.duplicate()

            self.current_step = self.input_data.scanning_data.get_additional_parameter("current_step")
            self.total_steps  = self.input_data.scanning_data.get_additional_parameter("total_steps")
            self.energy_step  = self.input_data.scanning_data.get_additional_parameter("photon_energy_step")

            self.total_power = self.input_data.scanning_data.get_additional_parameter("total_power")

            if self.cumulated_quantity == 1:  # Intensity
                self.total_power /= (1e3 * self.energy_step * codata.e)  # to ph/s

            self.energy_max = self.input_data.scanning_data.scanned_variable_value

            if self.energy_min is None:
                self.energy_min = self.input_data.scanning_data.scanned_variable_value
                self.cumulated_total_power = self.total_power
            else:
                self.cumulated_total_power += self.total_power

            if self.input_data.scanning_data.has_additional_parameter("is_footprint"):
                if self.input_data.scanning_data.get_additional_parameter("is_footprint"):
                    self.cb_rays.setEnabled(False)
                    self.rays = 0  # transmitted, absorbed doesn't make sense since is precalculated by footprint object
                else:
                    self.cb_rays.setEnabled(True)

            return True
        else:
            return False

    #########################################################
    # PLOTTING

    def replace_fig(self, shadow_data, var_x, var_y, xrange, yrange, nbins_h, nbins_v, nolost):
        if self.plot_canvas is None:
            self.plot_canvas = PowerPlotXYWidget()
            self.image_box.layout().addWidget(self.plot_canvas)
        try:
            if self.autosave == 1:
                if self.autosave_file is None:
                    self.autosave_file = ShadowPlot.PlotXYHdf5File(congruence.checkDir(self.autosave_file_name))
                elif self.autosave_file.filename != congruence.checkFileName(self.autosave_file_name):
                    self.autosave_file.close()
                    self.autosave_file = ShadowPlot.PlotXYHdf5File(congruence.checkDir(self.autosave_file_name))

                self.autosave_file.add_attribute("current_step", self.current_step, dataset_name="additional_data")
                self.autosave_file.add_attribute("total_steps", self.total_steps, dataset_name="additional_data")
                self.autosave_file.add_attribute("last_energy_value", self.energy_max, dataset_name="additional_data")
                self.autosave_file.add_attribute("last_power_value", self.total_power, dataset_name="additional_data")

            if self.keep_result == 1:
                self.cumulated_ticket, last_ticket = self.plot_canvas.plot_power_density(shadow_data, var_x, var_y,
                                                                                         self.total_power, self.cumulated_total_power,
                                                                                         self.energy_min, self.energy_max, self.energy_step,
                                                                                         nbins_h=nbins_h, nbins_v=nbins_v, xrange=xrange, yrange=yrange, nolost=nolost,
                                                                                         ticket_to_add=self.cumulated_ticket,
                                                                                         show_image=self.view_type==1,
                                                                                         kind_of_calculation=self.kind_of_calculation,
                                                                                         replace_poor_statistic=self.replace_poor_statistic,
                                                                                         good_rays_limit=self.good_rays_limit,
                                                                                         center_x=self.center_x,
                                                                                         center_y=self.center_y,
                                                                                         sigma_x=self.sigma_x,
                                                                                         sigma_y=self.sigma_y,
                                                                                         gamma=self.gamma,
                                                                                         cumulated_quantity=self.cumulated_quantity)

                if self.autosave == 1:
                    self.autosave_file.add_attribute("last_plotted_power",  self.cumulated_ticket['plotted_power'],  dataset_name="additional_data")
                    self.autosave_file.add_attribute("last_incident_power", self.cumulated_ticket['incident_power'], dataset_name="additional_data")
                    self.autosave_file.add_attribute("last_total_power",    self.cumulated_ticket['total_power'],    dataset_name="additional_data")
                    self.autosave_file.add_attribute("last_energy_min",     self.cumulated_ticket['energy_min'],     dataset_name="additional_data")
                    self.autosave_file.add_attribute("last_energy_max",     self.cumulated_ticket['energy_max'],     dataset_name="additional_data")
                    self.autosave_file.add_attribute("last_energy_step",    self.cumulated_ticket['energy_step'],    dataset_name="additional_data")

                self.plotted_ticket          = self.cumulated_ticket
                self.plotted_ticket_original = self.plotted_ticket.copy()

                if self.autosave == 1:
                    self.autosave_file.write_coordinates(self.cumulated_ticket)
                    dataset_name = "power_density"

                    self.autosave_file.add_plot_xy(self.cumulated_ticket, dataset_name=dataset_name)

                    if self.autosave_partial_results == 1:
                        if last_ticket is None:
                            self.autosave_file.add_plot_xy(self.cumulated_ticket,
                                                           plot_name="Energy Range: " + str(round(self.energy_max-self.energy_step, 2)) + "-" + str(round(self.energy_max, 2)),
                                                           dataset_name=dataset_name)
                        else:
                            self.autosave_file.add_plot_xy(last_ticket,
                                                           plot_name="Energy Range: " + str(round(self.energy_max-self.energy_step, 2)) + "-" + str(round(self.energy_max, 2)),
                                                           dataset_name=dataset_name)

                    self.autosave_file.flush()
            else:
                ticket, _ = self.plot_canvas.plot_power_density(shadow_data, var_x, var_y,
                                                                self.total_power, self.cumulated_total_power,
                                                                self.energy_min, self.energy_max, self.energy_step,
                                                                nbins_h=nbins_h, nbins_v=nbins_v, xrange=xrange, yrange=yrange, nolost=nolost,
                                                                show_image=self.view_type==1,
                                                                kind_of_calculation=self.kind_of_calculation,
                                                                replace_poor_statistic=self.replace_poor_statistic,
                                                                good_rays_limit=self.good_rays_limit,
                                                                center_x=self.center_x,
                                                                center_y=self.center_y,
                                                                sigma_x=self.sigma_x,
                                                                sigma_y=self.sigma_y,
                                                                gamma=self.gamma,
                                                                cumulated_quantity=self.cumulated_quantity)

                self.cumulated_ticket = None
                self.plotted_ticket = ticket
                self.plotted_ticket_original = self.plotted_ticket.copy()

                if self.autosave == 1:
                    self.autosave_file.write_coordinates(ticket)
                    self.autosave_file.add_plot_xy(ticket, dataset_name="power_density")
                    self.autosave_file.flush()

        except Exception as e:
            if not self.IS_DEVELOP:
                raise Exception("Data not plottable: " + str(e))
            else:
                raise e

add_widget_parameters_to_module(__name__)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# #########################################################################
# Copyright (c) 2020, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2020. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# #########################################################################

import sys, numpy, os

from AnyQt.QtWidgets import QMessageBox

from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Input, Output

from oasys2.canvas.util.canvas_util import add_widget_parameters_to_module

from oasys2.widget.util.exchange import DataExchangeObject
from oasys2.widget import gui as oasysgui
from oasys2.widget.util import congruence
from oasys2.widget.widget import OWWidget, OWAction

class TotalFilterCalculator(OWWidget):

    name = "Total Filter Calculator"
    description = "Total Filter Calculator"
    icon = "icons/total_filter.png"
    maintainer = "Luca Rebuffi"
    maintainer_email = "lrebuffi(@at@)anl.gov"
    priority = 6.1
    category = "User Defined"
    keywords = ["data", "file", "load", "read"]

    class Inputs:
        ref_1 = Input("Mirror Reflectivity #1", DataExchangeObject, default=True, auto_summary=False)
        ref_2 = Input("Mirror Reflectivity #2", DataExchangeObject, default=True, auto_summary=False)
        ref_3 = Input("Mirror Reflectivity #3", DataExchangeObject, default=True, auto_summary=False)
        dif_1 = Input("Diffraction Profile #1", DataExchangeObject, default=True, auto_summary=False)
        dif_2 = Input("Diffraction Profile #2", DataExchangeObject, default=True, auto_summary=False)
        mul_1 = Input("MultiLayer Reflectivity #1", DataExchangeObject, default=True, auto_summary=False)
        mul_2 = Input("MultiLayer Reflectivity #2", DataExchangeObject, default=True, auto_summary=False)
        pow_1 = Input("Power Output #1", DataExchangeObject, default=True, auto_summary=False)

    class Outputs:
        total_filter = Output("Total Filter", DataExchangeObject, default=True, auto_summary=False)

    want_main_area = 1

    ref_1 = None
    ref_2 = None
    ref_3 = None
    dif_1 = None
    dif_2 = None
    mul_1 = None
    mul_2 = None
    pow_1 = None

    check_ref_1 = Setting(0)
    check_ref_2 = Setting(0)
    check_ref_3 = Setting(0)
    check_dif_1 = Setting(0)
    check_dif_2 = Setting(0)
    check_mul_1 = Setting(0)
    check_mul_2 = Setting(0)
    check_pow_1 = Setting(0)

    n_ref_1 = Setting(0)
    n_ref_2 = Setting(0)
    n_ref_3 = Setting(0)
    n_dif_1 = Setting(0)
    n_dif_2 = Setting(0)
    n_mul_1 = Setting(0)
    n_mul_2 = Setting(0)
    abs_tran = Setting(1)

    energy_from = Setting(1000.0)
    energy_to   = Setting(51000.0)
    energy_nr = Setting(5000)

    CONTROL_AREA_WIDTH = 405

    def __init__(self):
        super(TotalFilterCalculator, self).__init__()

        self.runaction = OWAction("Calculate Total Filter", self)
        self.runaction.triggered.connect(self.calculate_total_filter)
        self.addAction(self.runaction)

        self.setFixedWidth(1200)
        self.setFixedHeight(710)

        self.controlArea.setFixedWidth(self.CONTROL_AREA_WIDTH)

        box0 = oasysgui.widgetBox(self.controlArea, "", orientation="horizontal", width=self.CONTROL_AREA_WIDTH-5)
        gui.button(box0, self, "Calculate Total Filter", callback=self.calculate_total_filter, height=45)

        box1 = oasysgui.widgetBox(self.controlArea, "Energy Range", orientation="vertical", width=self.CONTROL_AREA_WIDTH-5)

        oasysgui.lineEdit(box1, self, "energy_from", "Energy From [eV]", labelWidth=250, orientation="horizontal", valueType=float)
        oasysgui.lineEdit(box1, self, "energy_to", "Energy to [eV]", labelWidth=250, orientation="horizontal", valueType=float)
        oasysgui.lineEdit(box1, self, "energy_nr", "Nr. of energies", labelWidth=250, orientation="horizontal", valueType=int)

        box1 = oasysgui.widgetBox(self.controlArea, "Input Data", orientation="vertical", width=self.CONTROL_AREA_WIDTH-5)

        items = ["1", "2", "3", "4", "5", "6", "7", "8", "9", "10"]

        def build_combo(container, master, field_name, label, has_combo=True, combo_field_name=None, cb_items=items):
            box = gui.widgetBox(container, "", orientation="horizontal")
            gui.checkBox(box, master, "check_" + field_name, label="" if has_combo else label)
            if has_combo: gui.comboBox(box, master, "n_" + field_name if not combo_field_name else combo_field_name,
                                       label=label, items=cb_items, labelWidth=350, sendSelectedValue=False, orientation="horizontal")
            box.setEnabled(False)

            return box

        self.cb_ref_1 = build_combo(box1, self, "ref_1", label="Nr. of Mirror Reflectivity #1")
        self.cb_ref_2 = build_combo(box1, self, "ref_2", label="Nr. of Mirror Reflectivity #2")
        self.cb_ref_3 = build_combo(box1, self, "ref_3", label="Nr. of Mirror Reflectivity #3")

        gui.separator(box1)

        self.cb_dif_1 = build_combo(box1, self, "dif_1", label="Nr. of Diffraction Profile #1")
        self.cb_dif_2 = build_combo(box1, self, "dif_2", label="Nr. of Diffraction Profile #2")

        gui.separator(box1)

        self.cb_mul_1 = build_combo(box1, self, "mul_1", label="Nr. of MultiLayer Reflectivity #1")
        self.cb_mul_2 = build_combo(box1, self, "mul_2", label="Nr. of MultiLayer Reflectivity #2")

        gui.separator(box1)

        self.cb_pow_1 = build_combo(box1, self, "pow_1", label="Power Widget Input", combo_field_name="abs_tran", cb_items=["Absorbed", "Transmitted"])

        gui.rubber(self.controlArea)

        tabs = oasysgui.tabWidget(self.mainArea)
        tabs.setFixedHeight(self.height()-15)
        tabs.setFixedWidth(770)

        tab_fil = oasysgui.createTabPage(tabs, "Total Filter")

        self.filter_plot = oasysgui.plotWindow(tab_fil, position=True)
        self.filter_plot.setFixedHeight(self.height()-60)
        self.filter_plot.setFixedWidth(760)
        self.filter_plot.setGraphXLabel("Energy [eV]")
        self.filter_plot.setGraphYLabel("Intensity Factor")
        self.filter_plot.setGraphTitle("Total Filter")


    def __get_mirror_reflectivity(self, exchange_data):
        try:
            if not exchange_data.get_program_name() == "XOPPY": raise ValueError("Only XOPPY widgets are accepted")
            if not exchange_data.get_widget_name() == "XF1F2": raise ValueError("Only XF1F2 widgets are accepted")

            return exchange_data.get_content("xoppy_data")
        except Exception as e:
            QMessageBox.critical(self, "Error", str(e), QMessageBox.Ok)

            if self.IS_DEVELOP: raise e

    def __get_diffraction_profile(self, exchange_data):
        try:
            if not exchange_data.get_program_name() == "XOPPY": raise ValueError("Only XOPPY widgets are accepted")
            if not exchange_data.get_widget_name() == "XCRYSTAL": raise ValueError("Only XCRYSTAL widgets are accepted")

            xoppy_data = exchange_data.get_content("xoppy_data")

            energies = xoppy_data[:, 0]

            reflectivity_s = xoppy_data[:, 5]
            reflectivity_p = xoppy_data[:, 6]

            reflectivity_s[numpy.where(numpy.isnan(reflectivity_s))] = 0.0
            reflectivity_p[numpy.where(numpy.isnan(reflectivity_p))] = 0.0

            data = numpy.zeros((len(energies), 2))
            data[:, 0] = energies
            data[:, 1] = 0.5 * (reflectivity_p + reflectivity_s)

            return data
        except Exception as e:
            QMessageBox.critical(self, "Error", str(e), QMessageBox.Ok)

            if self.IS_DEVELOP: raise e

    def __get_multilayer_reflectivity(self, exchange_data):
        try:
            if not exchange_data.get_program_name() == "XOPPY": raise ValueError("Only XOPPY widgets are accepted")
            if not exchange_data.get_widget_name() == "MULTILAYER": raise ValueError("Only MULTILAYER widgets are accepted")

            xoppy_data = exchange_data.get_content("xoppy_data")

            energies = xoppy_data[:, 0]

            reflectivity_s = xoppy_data[:, 1]
            reflectivity_p = xoppy_data[:, 2]

            reflectivity_s[numpy.where(numpy.isnan(reflectivity_s))] = 0
            reflectivity_p[numpy.where(numpy.isnan(reflectivity_p))] = 0

            data = numpy.zeros((len(energies), 2))
            data[:, 0] = energies
            data[:, 1] = 0.5 * (reflectivity_p + reflectivity_s)

            return data
        except Exception as e:
            QMessageBox.critical(self, "Error", str(e), QMessageBox.Ok)

            if self.IS_DEVELOP: raise e

    def __get_power_data(self, exchange_data):
        try:
            if not exchange_data.get_program_name() == "XOPPY": raise ValueError("Only XOPPY widgets are accepted")
            if not exchange_data.get_widget_name() == "POWER": raise ValueError("Only POWER widgets are accepted")

            xoppy_data = exchange_data.get_content("xoppy_data")

            energies = xoppy_data[:, 0]

            data = numpy.zeros((len(energies), 2))
            data[:, 0] = energies

            data[:, 1] = xoppy_data[:, -1] / xoppy_data[:, 1]

            return data
        except Exception as e:
            QMessageBox.critical(self, "Error", str(e), QMessageBox.Ok)

            if self.IS_DEVELOP: raise e

    def __set_data(self, exchange_data, getter_method, field, combobox):
        if exchange_data is None: 
            setattr(self, field, None)
            setattr(self, "check_" + field, 0)
            combobox.setEnabled(False)
        else:
            try:
                setattr(self, field, getter_method(exchange_data))
                setattr(self, "check_" + field, 1)
                combobox.setEnabled(True)
            except Exception as e:
                combobox.setEnabled(False)
                
                QMessageBox.critical(self, "Error", str(e), QMessageBox.Ok)
                
                if self.IS_DEVELOP: raise e

    @Inputs.ref_1
    def set_ref_1(self, exchange_data): self.__set_data(exchange_data, self.__get_mirror_reflectivity,     "ref_1", self.cb_ref_1)
    @Inputs.ref_2
    def set_ref_2(self, exchange_data): self.__set_data(exchange_data, self.__get_mirror_reflectivity,     "ref_2", self.cb_ref_2)
    @Inputs.ref_3
    def set_ref_3(self, exchange_data): self.__set_data(exchange_data, self.__get_mirror_reflectivity,     "ref_3", self.cb_ref_3)
    @Inputs.dif_1
    def set_dif_1(self, exchange_data): self.__set_data(exchange_data, self.__get_diffraction_profile,     "dif_1", self.cb_dif_1)
    @Inputs.dif_2
    def set_dif_2(self, exchange_data): self.__set_data(exchange_data, self.__get_diffraction_profile,     "dif_2", self.cb_dif_2)
    @Inputs.mul_1
    def set_mul_1(self, exchange_data): self.__set_data(exchange_data, self.__get_multilayer_reflectivity, "mul_1", self.cb_mul_1)
    @Inputs.mul_2
    def set_mul_2(self, exchange_data): self.__set_data(exchange_data, self.__get_multilayer_reflectivity, "mul_2", self.cb_mul_2)
    @Inputs.pow_1
    def set_pow_1(self, exchange_data): self.__set_data(exchange_data, self.__get_power_data,              "pow_1", self.cb_pow_1)

    def __add_data_to_total_filter(self, total_filter, data, nr_data, check, absorbed=False):
        if not data is None and check==1:
            if absorbed:
                total_filter[:, 1] *= numpy.interp(total_filter[:, 0], data[:, 0], 1 - data[:, 1]) ** (nr_data + 1)
            else:
                total_filter[:, 1] *= numpy.interp(total_filter[:, 0], data[:, 0], data[:, 1]) ** (nr_data + 1)

    def calculate_total_filter(self):
        try:
            congruence.checkStrictlyPositiveNumber(self.energy_from, "Energy From")
            congruence.checkStrictlyPositiveNumber(self.energy_to, "Energy to")
            congruence.checkStrictlyPositiveNumber(self.energy_nr, "Nr. of energies")
            congruence.checkLessThan(self.energy_from, self.energy_to, "Energy From", "Energy to")

            energies = numpy.linspace(self.energy_from, self.energy_to, self.energy_nr)

            total_filter = numpy.ones((self.energy_nr, 2))
            total_filter[:, 0] = energies

            self.__add_data_to_total_filter(total_filter, self.ref_1, self.n_ref_1, self.check_ref_1)
            self.__add_data_to_total_filter(total_filter, self.ref_2, self.n_ref_2, self.check_ref_2)
            self.__add_data_to_total_filter(total_filter, self.ref_3, self.n_ref_3, self.check_ref_3)

            self.__add_data_to_total_filter(total_filter, self.dif_1, self.n_dif_1, self.check_dif_1)
            self.__add_data_to_total_filter(total_filter, self.dif_2, self.n_dif_2, self.check_dif_2)

            self.__add_data_to_total_filter(total_filter, self.mul_1, self.n_mul_1, self.check_mul_1)
            self.__add_data_to_total_filter(total_filter, self.mul_2, self.n_mul_2, self.check_mul_2)

            self.__add_data_to_total_filter(total_filter, self.pow_1, 0, self.check_pow_1, absorbed=self.abs_tran==0)

            total_filter[numpy.where(numpy.isnan(total_filter))] = 0.0

            if not total_filter[:, 1].sum() == total_filter.shape[0]:
                self.filter_plot.clear()
                self.filter_plot.addCurve(total_filter[:, 0], total_filter[:, 1], replace=True, legend="Total Filter")
                self.filter_plot.setGraphXLabel("Energy [eV]")
                self.filter_plot.setGraphYLabel("Intensity Factor")
                self.filter_plot.setGraphTitle("Total Filter")

                calculated_data = DataExchangeObject("Shadow4-Thermal", "TOTAL_FILTER")
                calculated_data.add_content("total_filter", total_filter)

                self.Outputs.total_filter.send(calculated_data)
            else:
                raise ValueError("Calculation not possibile: no input data")
        except Exception as e:
            QMessageBox.critical(self, "Error", str(e), QMessageBox.Ok)

            if self.IS_DEVELOP: raise e

add_widget_parameters_to_module(__name__)
